from typing import List, Optional

from cryptography.x509 import load_der_x509_certificate
from OpenSSL.crypto import X509, X509Store, X509StoreContext, X509StoreContextError

from .exceptions import InvalidCertificateChain
from .pem_cert_bytes_to_open_ssl_x509 import pem_cert_bytes_to_open_ssl_x509


def validate_certificate_chain(
    *,
    x5c: List[bytes],
    pem_root_certs_bytes: Optional[List[bytes]] = None,
) -> bool:
    """Validate that the certificates in x5c chain back to a known root certificate

    Args:
        `x5c`: X5C certificates from a registration response's attestation statement
        (optional) `pem_root_certs_bytes`: Any additional (PEM-formatted)
        root certificates that may complete the certificate chain

    Raises:
        `helpers.exceptions.InvalidCertificateChain` if chain cannot be validated
    """
    if pem_root_certs_bytes is None or len(pem_root_certs_bytes) < 1:
        # We have no root certs to chain back to, so just pass on validation
        return True

    # Make sure we have at least one certificate to try and link back to a root cert
    if len(x5c) < 1:
        raise InvalidCertificateChain("x5c was empty")

    # Prepare leaf cert
    try:
        leaf_cert_bytes = x5c[0]
        leaf_cert_crypto = load_der_x509_certificate(leaf_cert_bytes)
        leaf_cert = X509().from_cryptography(leaf_cert_crypto)
    except Exception as exc:
        raise InvalidCertificateChain("Could not prepare leaf cert") from exc

    # Prepare any intermediate certs
    try:
        # May be an empty array, that's fine
        intermediate_certs_bytes = x5c[1:]
        intermediate_certs_crypto = [
            load_der_x509_certificate(cert) for cert in intermediate_certs_bytes
        ]
        intermediate_certs = [X509().from_cryptography(cert) for cert in intermediate_certs_crypto]
    except Exception as exc:
        raise InvalidCertificateChain("Could not prepare intermediate certs") from exc

    # Prepare a collection of possible root certificates
    cert_store = _generate_new_cert_store()
    try:
        for cert in pem_root_certs_bytes:
            cert_store.add_cert(pem_cert_bytes_to_open_ssl_x509(cert))
    except Exception as exc:
        raise InvalidCertificateChain("Could not prepare root certs") from exc

    # Load certs into a "context" for validation
    context = X509StoreContext(
        store=cert_store,
        certificate=leaf_cert,
        chain=intermediate_certs,
    )

    # Validate the chain (will raise if it can't)
    try:
        context.verify_certificate()
    except X509StoreContextError as exc:
        raise InvalidCertificateChain("Certificate chain could not be validated") from exc

    return True


def _generate_new_cert_store() -> X509Store:
    """
    Something that can be patched during testing to return an X509Store instance with its time
    adjusted to some datetime in the past. This allows full certificate validity checks even after
    cert expiration in the real world. See here:

    https://www.pyopenssl.org/en/stable/api/crypto.html#OpenSSL.crypto.X509Store.set_time
    """
    return X509Store()



import secrets
from typing import Optional, Tuple
from unittest import TestCase
import cbor2

from webauthn.helpers import parse_authenticator_data, bytes_to_base64url, parse_cbor
from webauthn.helpers.base64url_to_bytes import base64url_to_bytes


def _generate_auth_data(
    sign_count: int = 0,
    up: bool = True,
    uv: bool = False,
    be: bool = False,
    bs: bool = False,
    at: bool = False,
    ed: bool = False,
) -> Tuple[bytes, bytes, int, Optional[bytes], Optional[bytes], Optional[bytes]]:
    """A helper to generate auth_data

    Args:
        `sign_count`: How many times the authenticator has been used
        `up`: Whether user was present
        `uv`: Whether user was verified
        `be`: Whether credential can be backed up
        `bs`: Whether credential has been backed up
        `at`: Whether attested credential data is present
        `ed`: Whether extension data is present

    Returns:
        A `tuple` comprised of the following values:
            `bytes`: Authenticator data
            `bytes`: RP ID hash
            `int`: Sign count
            `Optional[bytes]`: AAGUID
            `Optional[bytes]`: Credential ID
            `Optional[bytes]`: Credential public key
    """
    rp_id_hash = secrets.token_bytes(32)

    flags = 0b00000000
    if up is True:
        flags = flags | 1 << 0
    if uv is True:
        flags = flags | 1 << 2
    if be is True:
        flags = flags | 1 << 3
    if bs is True:
        flags = flags | 1 << 4
    if at is True:
        flags = flags | 1 << 6
    if ed is True:
        flags = flags | 1 << 7

    bytes_to_join = [
        rp_id_hash,
        flags.to_bytes(1, byteorder="big"),
        sign_count.to_bytes(4, byteorder="big"),
    ]

    aaguid: Optional[bytes] = None
    credential_id: Optional[bytes] = None
    credential_public_key: Optional[bytes] = None
    if at is True:
        aaguid = secrets.token_bytes(16)
        credential_id = secrets.token_bytes(32)
        credential_public_key = secrets.token_bytes(32)

        attested_data = [
            aaguid,
            len(credential_id).to_bytes(2, byteorder="big"),
            credential_id,
            credential_public_key,
        ]

        bytes_to_join += attested_data

    auth_data = b"".join(bytes_to_join)

    return (
        auth_data,
        rp_id_hash,
        sign_count,
        aaguid,
        credential_id,
        credential_public_key,
    )


class TestWebAuthnParseAuthenticatorData(TestCase):
    def test_correctly_parses_simple(self) -> None:
        (auth_data, rp_id_hash, sign_count, _, _, _) = _generate_auth_data(10, up=True, uv=True)

        output = parse_authenticator_data(auth_data)

        assert output.rp_id_hash == rp_id_hash
        assert output.flags.up is True
        assert output.flags.uv is True
        assert output.flags.be is False
        assert output.flags.be is False
        assert output.flags.at is False
        assert output.flags.ed is False
        assert output.sign_count == sign_count

    def test_correctly_parses_attested_credential_data(self) -> None:
        auth_data = base64url_to_bytes(
            "SZYN5YgOjGh0NBcPZHZgW4_krrmihjLHmVzzuoMdl2NBAAAAJch83ZdWwUm4niTLNjZU81AAIHa7Ksm5br3hAh3UjxP9-4rqu8BEsD-7SZ2xWe1_yHv6pAEDAzkBACBZAQDcxA7Ehs9goWB2Hbl6e9v-aUub9rvy2M7Hkvf-iCzMGE63e3sCEW5Ru33KNy4um46s9jalcBHtZgtEnyeRoQvszis-ws5o4Da0vQfuzlpBmjWT1dV6LuP-vs9wrfObW4jlA5bKEIhv63-jAxOtdXGVzo75PxBlqxrmrr5IR9n8Fw7clwRsDkjgRHaNcQVbwq_qdNwU5H3hZKu9szTwBS5NGRq01EaDF2014YSTFjwtAmZ3PU1tcO_QD2U2zg6eB5grfWDeAJtRE8cbndDWc8aLL0aeC37Q36-TVsGe6AhBgHEw6eO3I3NW5r9v_26CqMPBDwmEundeq1iGyKfMloobIUMBAAE"
        )

        output = parse_authenticator_data(auth_data)

        cred_data = output.attested_credential_data
        self.assertIsNotNone(cred_data)
        assert cred_data  # Make mypy happy
        self.assertEqual(
            cred_data.aaguid,
            base64url_to_bytes("yHzdl1bBSbieJMs2NlTzUA"),
        )
        self.assertEqual(
            cred_data.credential_id,
            base64url_to_bytes("drsqybluveECHdSPE_37iuq7wESwP7tJnbFZ7X_Ie_o"),
        )
        self.assertEqual(
            cred_data.credential_public_key,
            base64url_to_bytes(
                "pAEDAzkBACBZAQDcxA7Ehs9goWB2Hbl6e9v-aUub9rvy2M7Hkvf-iCzMGE63e3sCEW5Ru33KNy4um46s9jalcBHtZgtEnyeRoQvszis-ws5o4Da0vQfuzlpBmjWT1dV6LuP-vs9wrfObW4jlA5bKEIhv63-jAxOtdXGVzo75PxBlqxrmrr5IR9n8Fw7clwRsDkjgRHaNcQVbwq_qdNwU5H3hZKu9szTwBS5NGRq01EaDF2014YSTFjwtAmZ3PU1tcO_QD2U2zg6eB5grfWDeAJtRE8cbndDWc8aLL0aeC37Q36-TVsGe6AhBgHEw6eO3I3NW5r9v_26CqMPBDwmEundeq1iGyKfMloobIUMBAAE"
            ),
        )

    def test_parses_uv_false(self) -> None:
        auth_data = _generate_auth_data()[0]

        output = parse_authenticator_data(auth_data)

        self.assertTrue(output.flags.up)
        self.assertFalse(output.flags.uv)

    def test_parses_attested_credential_data_and_extension_data(self) -> None:
        auth_data = bytes.fromhex(
            "50569158be61d7a1ba084f80e45e938fd326e0a8dff07b37036e6c82303ae26bc1000004377b3024675546afcb92e4495c8a1e193f00dca30058b8d74f6bd74de90baeb34afb51e3578e1ac4ca9f79a7f88473d8254d5762ca82d68f3bf63f49e9b284caab4d45d6f9bb468d0c1b7f0f727378c1db8adb4802cb7c5ad9c5eb905bf0ba03f79bd1f04d63765452d49c4087acfad340516dc892eafd87d498ae9e6fd6f06a3f423108ebdc032d93e82fdd6deacc1b638fd56838a482f01232ad01e266e016a50b8121816997a167f41139900fe46094b8ef30aad14ee08cc457366a033bb4a0554dcf9c9589f9622d4f84481541014c870291c87d7a3bbe3d8b07eb02509de5721e3f728aa5eac41e9c5af02869a4010103272006215820e613b86a8d4ebae24e84a0270b6773f7bb30d1d59f5ec379910ebe7c87714274a16b6372656450726f7465637401"
        )
        output = parse_authenticator_data(auth_data)

        cred_data = output.attested_credential_data
        self.assertIsNotNone(cred_data)
        assert cred_data  # Make mypy happy
        self.assertEqual(
            bytes_to_base64url(cred_data.credential_public_key),
            "pAEBAycgBiFYIOYTuGqNTrriToSgJwtnc_e7MNHVn17DeZEOvnyHcUJ0",
        )

        extensions = output.extensions
        self.assertIsNotNone(extensions)
        assert extensions  # Make mypy happy

        parsed_extensions = parse_cbor(extensions)
        self.assertEqual(parsed_extensions, {"credProtect": 1})

    def test_parses_only_extension_data(self) -> None:
        # Pulled from Conformance Testing suite
        auth_data = base64url_to_bytes(
            "SZYN5YgOjGh0NBcPZHZgW4_krrmihjLHmVzzuoMdl2OBAAAAjaFxZXhhbXBsZS5leHRlbnNpb254dlRoaXMgaXMgYW4gZXhhbXBsZSBleHRlbnNpb24hIElmIHlvdSByZWFkIHRoaXMgbWVzc2FnZSwgeW91IHByb2JhYmx5IHN1Y2Nlc3NmdWxseSBwYXNzaW5nIGNvbmZvcm1hbmNlIHRlc3RzLiBHb29kIGpvYiE"
        )

        output = parse_authenticator_data(auth_data)

        extensions = output.extensions
        self.assertIsNotNone(extensions)
        assert extensions  # Make mypy happy
        parsed_extensions = parse_cbor(extensions)
        self.assertEqual(
            parsed_extensions,
            {
                "example.extension": "This is an example extension! If you read this message, you probably successfully passing conformance tests. Good job!",
            },
        )

    def test_parses_backup_state_flags(self) -> None:
        (auth_data, _, _, _, _, _) = _generate_auth_data(be=True, bs=True)

        output = parse_authenticator_data(auth_data)

        assert output.flags.be is True
        assert output.flags.be is True

    def test_parses_bad_eddsa_auth_data(self) -> None:
        """
        Help out particular YubiKeys that incorrectly CBOR-encode authData when they use Ed25519
        for their public key.

        See https://github.com/duo-labs/py_webauthn/issues/160
        """
        auth_data = bytearray.fromhex(
            "16B02DFBC3D4CCA37EBC2F6516659B12210DB9E1018AB9F13A9690638EA6FDA845000000012FC0579F811347EAB116BB5A8DB9202A0080E82FE6BDE300E4ECC93E0016448AD00FA6F28A011A6F87FF7B0CFCA499BEAF83344C3660B5ECABF72A3B2838A0CC7D87D3FA58292B53449CFF13AD69732D7521649D365CCBC5D0A0FA4B4E09EAE99537261F2F44093F8F4FD4CF5796E0FE58FF0615FFC5882836BBE7B99B08BE2986721C1C5A6AC7F32D3220D9B34D8DEE2FC9A301634F4B5003272067456432353531392198201618F6185918FA182E141875183A18841718521874187A18C51318D918C51883182D18ED181818EA188F182E187407185E18F41518CC18C9186D"
        )

        output = parse_authenticator_data(auth_data)

        cred_data = output.attested_credential_data
        self.assertIsNotNone(cred_data)
        assert cred_data  # Make mypy happy

        self.assertEqual(
            cred_data.credential_id.hex(),
            "e82fe6bde300e4ecc93e0016448ad00fa6f28a011a6f87ff7b0cfca499beaf83344c3660b5ecabf72a3b2838a0cc7d87d3fa58292b53449cff13ad69732d7521649d365ccbc5d0a0fa4b4e09eae99537261f2f44093f8f4fd4cf5796e0fe58ff0615ffc5882836bbe7b99b08be2986721c1c5a6ac7f32d3220d9b34d8dee2fc9",
        )
        self.assertEqual(
            cred_data.credential_public_key.hex(),
            "a401634f4b5003272067456432353531392198201618f6185918fa182e141875183a18841718521874187a18c51318d918c51883182d18ed181818ea188f182e187407185e18f41518cc18c9186d",
        )



from unittest import TestCase

from webauthn.helpers import base64url_to_bytes, bytes_to_base64url
from webauthn.helpers.cose import COSEKTY, COSEAlgorithmIdentifier
from webauthn.helpers.decode_credential_public_key import (
    DecodedEC2PublicKey,
    DecodedRSAPublicKey,
    decode_credential_public_key,
)


class TestDecodeCredentialPublicKey(TestCase):
    def test_decodes_ec2_public_key(self) -> None:
        decoded = decode_credential_public_key(
            base64url_to_bytes(
                "pQECAyYgASFYIDDHBDxTqWP4yZZnAa524L6uPuwhireUwRD5sXY6U2gxIlggxuwbECbDdNfTTegnc174oYdusZiMmJgct0yI_ulrJGI"
            )
        )

        assert isinstance(decoded, DecodedEC2PublicKey)
        assert decoded.kty == COSEKTY.EC2
        assert decoded.alg == COSEAlgorithmIdentifier.ECDSA_SHA_256
        assert decoded.crv == 1
        assert (
            decoded.x
            and bytes_to_base64url(decoded.x) == "MMcEPFOpY_jJlmcBrnbgvq4-7CGKt5TBEPmxdjpTaDE"
        )
        assert (
            decoded.y
            and bytes_to_base64url(decoded.y) == "xuwbECbDdNfTTegnc174oYdusZiMmJgct0yI_ulrJGI"
        )

    def test_decode_rsa_public_key(self) -> None:
        decoded = decode_credential_public_key(
            base64url_to_bytes(
                "pAEDAzkBACBZAQDxfpXrj0ba_AH30JJ_-W7BHSOPugOD8aEDdNBKc1gjB9AmV3FPl2aL0fwiOMKtM_byI24qXb2FzcyjC7HUVkHRtzkAQnahXckI4wY_01koaY6iwXuIE3Ya0Zjs2iZyz6u4G_abGnWdObqa_kHxc3CHR7Xy5MDkAkKyX6TqU0tgHZcEhDd_Lb5ONJDwg4wvKlZBtZYElfMuZ6lonoRZ7qR_81rGkDZyFaxp6RlyvzEbo4ijeIaHQylqCz-oFm03ifZMOfRHYuF4uTjJDRH-g4BW1f3rdi7DTHk1hJnIw1IyL_VFIQ9NifkAguYjNCySCUNpYli2eMrPhAu5dYJFFjINIUMBAAE"
            )
        )

        assert isinstance(decoded, DecodedRSAPublicKey)
        assert decoded.kty == COSEKTY.RSA
        assert decoded.alg == COSEAlgorithmIdentifier.RSASSA_PKCS1_v1_5_SHA_256
        assert decoded.e and bytes_to_base64url(decoded.e) == "AQAB"
        assert (
            decoded.n
            and bytes_to_base64url(decoded.n)
            == "8X6V649G2vwB99CSf_luwR0jj7oDg_GhA3TQSnNYIwfQJldxT5dmi9H8IjjCrTP28iNuKl29hc3Mowux1FZB0bc5AEJ2oV3JCOMGP9NZKGmOosF7iBN2GtGY7Nomcs-ruBv2mxp1nTm6mv5B8XNwh0e18uTA5AJCsl-k6lNLYB2XBIQ3fy2-TjSQ8IOMLypWQbWWBJXzLmepaJ6EWe6kf_NaxpA2chWsaekZcr8xG6OIo3iGh0Mpags_qBZtN4n2TDn0R2LheLk4yQ0R_oOAVtX963Yuw0x5NYSZyMNSMi_1RSEPTYn5AILmIzQskglDaWJYtnjKz4QLuXWCRRYyDQ"
        )

    def test_decode_uncompressed_ec2_public_key(self) -> None:
        decoded = decode_credential_public_key(
            base64url_to_bytes(
                "BBaxKZueVyr5ICDfosygxwRflSdPUcNheZhThXCeTFTNo0EM9dj0V+xJ1JwpE2XZ/8NRIt5KVvr71Zl0rB8BWOs="
            )
        )

        assert isinstance(decoded, DecodedEC2PublicKey)
        assert decoded.kty == COSEKTY.EC2
        assert decoded.alg == COSEAlgorithmIdentifier.ECDSA_SHA_256
        assert decoded.crv == 1
        assert (
            decoded.x
            and bytes_to_base64url(decoded.x) == "FrEpm55XKvkgIN-izKDHBF-VJ09Rw2F5mFOFcJ5MVM0"
        )
        assert (
            decoded.y
            and bytes_to_base64url(decoded.y) == "o0EM9dj0V-xJ1JwpE2XZ_8NRIt5KVvr71Zl0rB8BWOs"
        )



from unittest import TestCase

from webauthn.helpers.tpm.parse_cert_info import parse_cert_info
from webauthn.helpers.tpm.structs import TPM_ALG, TPM_ST


class TestWebAuthnTPMParseCertInfo(TestCase):
    def test_properly_parses_cert_info_bytes(self) -> None:
        cert_info = b'\xffTCG\x80\x17\x00"\x00\x0bW"f{J5_9"\x15\tL\x01\xd5e\xbcr\xc6\xc9\x03\xbc#\xb5m\xee\xb5yI+j\xe6\xce\x00\x14`\x0bD(A\x99\xf3\xd3\x12I[\x04\x1f\xf4\xe7\xfb)\xc8\x02\x8f\x00\x00\x00\x00\x1a0)Z\x16\xb0}\xb1R\'s\xf8\x01\x97g1K\xfaf`T\x00"\x00\x0b\xe7\x1c"\x90\x07\xdeA\xe1w\xe0\xb3F\xe1\x07\x02\x8c\x16b\xe1\r\x9e\xb8\xae\xe7\xa95\xac\xf6\x1a\xedx\x89\x00"\x00\x0b\x7f\xe8\x84\xdaC\xa7\xc5?\xcept,\xa9\nA\x99\x93\xbc\x1f\x15\xcbs\x7f\xe0\x1a\x96u\xca\xe4\x8f\x86\x81'

        output = parse_cert_info(cert_info)

        assert output.magic == b"\xffTCG"
        assert output.type == TPM_ST.ATTEST_CERTIFY
        assert (
            output.qualified_signer
            == b'\x00\x0bW"f{J5_9"\x15\tL\x01\xd5e\xbcr\xc6\xc9\x03\xbc#\xb5m\xee\xb5yI+j\xe6\xce'
        )
        assert output.extra_data == b"`\x0bD(A\x99\xf3\xd3\x12I[\x04\x1f\xf4\xe7\xfb)\xc8\x02\x8f"
        assert output.firmware_version == b"\x97g1K\xfaf`T"
        # Attested
        assert output.attested.name_alg == TPM_ALG.SHA256
        assert output.attested.name_alg_bytes == b"\x00\x0b"
        assert (
            output.attested.name
            == b'\x00\x0b\xe7\x1c"\x90\x07\xdeA\xe1w\xe0\xb3F\xe1\x07\x02\x8c\x16b\xe1\r\x9e\xb8\xae\xe7\xa95\xac\xf6\x1a\xedx\x89'
        )
        assert (
            output.attested.qualified_name
            == b"\x00\x0b\x7f\xe8\x84\xdaC\xa7\xc5?\xcept,\xa9\nA\x99\x93\xbc\x1f\x15\xcbs\x7f\xe0\x1a\x96u\xca\xe4\x8f\x86\x81"
        )
        # Clock Info
        assert output.clock_info.clock == b"\x00\x00\x00\x00\x1a0)Z"
        assert output.clock_info.reset_count == 380665265
        assert output.clock_info.restart_count == 1378317304
        assert output.clock_info.safe is True



from unittest import TestCase
from unittest.mock import MagicMock, patch

from webauthn.helpers.cose import COSEAlgorithmIdentifier
from webauthn.helpers.structs import (
    AttestationConveyancePreference,
    AuthenticatorAttachment,
    AuthenticatorSelectionCriteria,
    PublicKeyCredentialDescriptor,
    PublicKeyCredentialParameters,
    PublicKeyCredentialRpEntity,
    PublicKeyCredentialUserEntity,
    ResidentKeyRequirement,
)
from webauthn import generate_registration_options


class TestGenerateRegistrationOptions(TestCase):
    @patch("secrets.token_bytes")
    def test_generates_options_with_defaults(self, token_bytes_mock: MagicMock) -> None:
        token_bytes_mock.return_value = b"12345"
        user_id = "ABAV6QWPBEY9WOTOA1A4".encode("utf-8")

        options = generate_registration_options(
            rp_id="example.com",
            rp_name="Example Co",
            user_id=user_id,
            user_name="lee",
        )

        assert options.rp == PublicKeyCredentialRpEntity(
            id="example.com",
            name="Example Co",
        )
        assert options.challenge == b"12345"
        assert options.user == PublicKeyCredentialUserEntity(
            id=user_id,
            name="lee",
            display_name="lee",
        )
        assert options.pub_key_cred_params[0] == PublicKeyCredentialParameters(
            type="public-key",
            alg=COSEAlgorithmIdentifier.ECDSA_SHA_256,
        )
        assert options.timeout == 60000
        assert options.exclude_credentials == []
        assert options.authenticator_selection is None
        assert options.attestation == AttestationConveyancePreference.NONE

    def test_generates_options_with_custom_values(self) -> None:
        user_id = "ABAV6QWPBEY9WOTOA1A4".encode("utf-8")

        options = generate_registration_options(
            rp_id="example.com",
            rp_name="Example Co",
            user_id=user_id,
            user_name="lee",
            user_display_name="Lee",
            attestation=AttestationConveyancePreference.DIRECT,
            authenticator_selection=AuthenticatorSelectionCriteria(
                authenticator_attachment=AuthenticatorAttachment.PLATFORM,
                resident_key=ResidentKeyRequirement.REQUIRED,
            ),
            challenge=b"1234567890",
            exclude_credentials=[
                PublicKeyCredentialDescriptor(id=b"1234567890"),
            ],
            supported_pub_key_algs=[COSEAlgorithmIdentifier.ECDSA_SHA_512],
            timeout=120000,
        )

        assert options.rp == PublicKeyCredentialRpEntity(id="example.com", name="Example Co")
        assert options.challenge == b"1234567890"
        assert options.user == PublicKeyCredentialUserEntity(
            id=user_id,
            name="lee",
            display_name="Lee",
        )
        assert options.pub_key_cred_params[0] == PublicKeyCredentialParameters(
            type="public-key",
            alg=COSEAlgorithmIdentifier.ECDSA_SHA_512,
        )
        assert options.timeout == 120000
        assert options.exclude_credentials == [PublicKeyCredentialDescriptor(id=b"1234567890")]
        assert options.authenticator_selection == AuthenticatorSelectionCriteria(
            authenticator_attachment=AuthenticatorAttachment.PLATFORM,
            resident_key=ResidentKeyRequirement.REQUIRED,
            require_resident_key=True,
        )
        assert options.attestation == AttestationConveyancePreference.DIRECT

    def test_raises_on_empty_rp_id(self) -> None:
        with self.assertRaisesRegex(ValueError, "rp_id"):
            generate_registration_options(
                rp_id="",
                rp_name="Example Co",
                user_name="blah",
            )

    def test_raises_on_empty_rp_name(self) -> None:
        with self.assertRaisesRegex(ValueError, "rp_name"):
            generate_registration_options(
                rp_id="example.com",
                rp_name="",
                user_name="blah",
            )

    @patch("secrets.token_bytes")
    def test_generated_random_id_on_empty_user_id(self, token_bytes_mock: MagicMock) -> None:
        token_bytes_mock.return_value = bytes([1, 2, 3, 4])

        options = generate_registration_options(
            rp_id="example.com",
            rp_name="Example Co",
            user_name="blah",
            user_id=None,
        )

        self.assertEqual(options.user.id, bytes([1, 2, 3, 4]))

    def test_raises_on_non_bytes_user_id(self) -> None:
        with self.assertRaisesRegex(ValueError, "user_id"):
            generate_registration_options(
                rp_id="example.com",
                rp_name="Example Co",
                user_name="hello",
                user_id="hello",  # type: ignore
            )

    def test_raises_on_empty_user_name(self) -> None:
        with self.assertRaisesRegex(ValueError, "user_name"):
            generate_registration_options(
                rp_id="example.com",
                rp_name="Example Co",
                user_name="",
            )



from unittest import TestCase

from webauthn.helpers import parse_registration_credential_json
from webauthn.helpers.structs import (
    AuthenticatorTransport,
    AuthenticatorAttachment,
)


class TestStructsRegistrationCredential(TestCase):
    def test_parse_registration_credential_json(self):
        """
        Check that we can properly parse some values that aren't really here-or-there for response
        verification, but can still be useful to RP's to fine-tune the WebAuthn experience.
        """
        parsed = parse_registration_credential_json(
            """{
                "id": "ZoIKP1JQvKdrYj1bTUPJ2eTUsbLeFkv-X5xJQNr4k6s",
                "rawId": "ZoIKP1JQvKdrYj1bTUPJ2eTUsbLeFkv-X5xJQNr4k6s",
                "response": {
                    "attestationObject": "o2NmbXRkbm9uZWdhdHRTdG10oGhhdXRoRGF0YVkBZ0mWDeWIDoxodDQXD2R2YFuP5K65ooYyx5lc87qDHZdjRQAAAAAAAAAAAAAAAAAAAAAAAAAAACBmggo_UlC8p2tiPVtNQ8nZ5NSxst4WS_5fnElA2viTq6QBAwM5AQAgWQEA31dtHqc70D_h7XHQ6V_nBs3Tscu91kBL7FOw56_VFiaKYRH6Z4KLr4J0S12hFJ_3fBxpKfxyMfK66ZMeAVbOl_wemY4S5Xs4yHSWy21Xm_dgWhLJjZ9R1tjfV49kDPHB_ssdvP7wo3_NmoUPYMgK-edgZ_ehttp_I6hUUCnVaTvn_m76b2j9yEPReSwl-wlGsabYG6INUhTuhSOqG-UpVVQdNJVV7GmIPHCA2cQpJBDZBohT4MBGme_feUgm4sgqVCWzKk6CzIKIz5AIVnspLbu05SulAVnSTB3NxTwCLNJR_9v9oSkvphiNbmQBVQH1tV_psyi9HM1Jtj9VJVKMeyFDAQAB",
                    "clientDataJSON": "eyJ0eXBlIjoid2ViYXV0aG4uY3JlYXRlIiwiY2hhbGxlbmdlIjoiQ2VUV29nbWcwY2NodWlZdUZydjhEWFhkTVpTSVFSVlpKT2dhX3hheVZWRWNCajBDdzN5NzN5aEQ0RmtHU2UtUnJQNmhQSkpBSW0zTFZpZW40aFhFTGciLCJvcmlnaW4iOiJodHRwOi8vbG9jYWxob3N0OjUwMDAiLCJjcm9zc09yaWdpbiI6ZmFsc2V9",
                    "transports": ["internal", "hybrid"]
                },
                "type": "public-key",
                "clientExtensionResults": {},
                "authenticatorAttachment": "platform"
            }"""
        )

        self.assertEqual(
            parsed.response.transports,
            [
                AuthenticatorTransport.INTERNAL,
                AuthenticatorTransport.HYBRID,
            ],
        )
        self.assertEqual(parsed.authenticator_attachment, AuthenticatorAttachment.PLATFORM)



from unittest import TestCase
from datetime import datetime

from OpenSSL.crypto import X509Store

from webauthn.helpers import base64url_to_bytes
from webauthn.helpers.structs import AttestationFormat
from webauthn import verify_registration_response

from .helpers.x509store import patch_validate_certificate_chain_x509store_getter


class TestVerifyRegistrationResponseApple(TestCase):
    @patch_validate_certificate_chain_x509store_getter
    def test_verify_attestation_apple_passkey(
        self,
        patched_x509store: X509Store,
    ) -> None:
        # Setting the time to something that satisfies all these:
        # (Leaf) 20210831230207Z <-> 20210903230207Z <- Earliest expiration
        # (Int.) 20200318183801Z <-> 20300313000000Z
        # (Root) 20200318182132Z <-> 20450315000000Z
        patched_x509store.set_time(datetime(2021, 9, 1, 0, 0, 0))

        credential = """{
            "id": "0yhsKG_gCzynIgNbvXWkqJKL8Uc",
            "rawId": "0yhsKG_gCzynIgNbvXWkqJKL8Uc",
            "response": {
                "attestationObject": "o2NmbXRlYXBwbGVnYXR0U3RtdKFjeDVjglkCRzCCAkMwggHJoAMCAQICBgF7o5kiITAKBggqhkjOPQQDAjBIMRwwGgYDVQQDDBNBcHBsZSBXZWJBdXRobiBDQSAxMRMwEQYDVQQKDApBcHBsZSBJbmMuMRMwEQYDVQQIDApDYWxpZm9ybmlhMB4XDTIxMDgzMTIzMDIwN1oXDTIxMDkwMzIzMDIwN1owgZExSTBHBgNVBAMMQGIxMGY3MThiYzVkZDc1ODg4NmExZDhjZmI1YjhiNjMxNzI5ZjRkN2U0YmEwNjlhYjBhOTkyYzFjMDg0NzhhZjkxGjAYBgNVBAsMEUFBQSBDZXJ0aWZpY2F0aW9uMRMwEQYDVQQKDApBcHBsZSBJbmMuMRMwEQYDVQQIDApDYWxpZm9ybmlhMFkwEwYHKoZIzj0CAQYIKoZIzj0DAQcDQgAE0SSw6f-BknI8nuL6T4Fw03PgMobPiAruxwCKFM3qZHJJY-BbuMRKn5gN7RKqijN5XPgdMedBFs7W8fTF6ww1j6NVMFMwDAYDVR0TAQH_BAIwADAOBgNVHQ8BAf8EBAMCBPAwMwYJKoZIhvdjZAgCBCYwJKEiBCDkV-W8KS8WNSECSO0ud2uhKcfMRpUkp1NWg2yu8vBYoDAKBggqhkjOPQQDAgNoADBlAjBlxucHXdrLUIeahBKQR1kBPQ2nhyZAh1mgHxmUwXlaacLB0RMGwtG8l75hQWJ7hncCMQCrC559l8orYDse224mTEm_GXE4DCr6XTf4xP9aXebUV6GcuAwCsu35SwhT4EgvhoZZAjgwggI0MIIBuqADAgECAhBWJVOVx6f7QOviKNgmCFO2MAoGCCqGSM49BAMDMEsxHzAdBgNVBAMMFkFwcGxlIFdlYkF1dGhuIFJvb3QgQ0ExEzARBgNVBAoMCkFwcGxlIEluYy4xEzARBgNVBAgMCkNhbGlmb3JuaWEwHhcNMjAwMzE4MTgzODAxWhcNMzAwMzEzMDAwMDAwWjBIMRwwGgYDVQQDDBNBcHBsZSBXZWJBdXRobiBDQSAxMRMwEQYDVQQKDApBcHBsZSBJbmMuMRMwEQYDVQQIDApDYWxpZm9ybmlhMHYwEAYHKoZIzj0CAQYFK4EEACIDYgAEgy6HLyYUkYECJbn1_Na7Y3i19V8_ywRbxzWZNHX9VJBE35v-GSEXZcaaHdoFCzjUUINAGkNPsk0RLVbD4c-_y5iR_sBpYIG--Wy8d8iN3a9Gpa7h3VFbWvqrk76cCyaRo2YwZDASBgNVHRMBAf8ECDAGAQH_AgEAMB8GA1UdIwQYMBaAFCbXZNnFeMJaZ9Gn3msS0Btj8cbXMB0GA1UdDgQWBBTrroLE_6GsW1HUzyRhBQC-Y713iDAOBgNVHQ8BAf8EBAMCAQYwCgYIKoZIzj0EAwMDaAAwZQIxAN2LGjSBpfrZ27TnZXuEHhRMJ7dbh2pBhsKxR1dQM3In7-VURX72SJUMYy5cSD5wwQIwLIpgRNwgH8_lm8NNKTDBSHhR2WDtanXx60rKvjjNJbiX0MgFvvDH94sHpXHG6A4HaGF1dGhEYXRhWJiPh6BZvowZk4E0cyGRAQ-e4LvoufWAcLD1j4UMTOIowUUAAAAA8kqOcNDT-CwpNzJSPMTeWgAU0yhsKG_gCzynIgNbvXWkqJKL8UelAQIDJiABIVgg0SSw6f-BknI8nuL6T4Fw03PgMobPiAruxwCKFM3qZHIiWCBJY-BbuMRKn5gN7RKqijN5XPgdMedBFs7W8fTF6ww1jw",
                "clientDataJSON": "eyJ0eXBlIjoid2ViYXV0aG4uY3JlYXRlIiwiY2hhbGxlbmdlIjoiMW5ocXlNa2ZHQVFMLXRUY3NmcHVveXE4aHFlb0hyMGQ5dERtanYxQnVKOTdZVEEzRkxXUzVFZFk0cVVnLU16cnVjMnNpQmR5VmxuRklQQjFnMEhoMkEiLCJvcmlnaW4iOiJodHRwczovL2RldjIuZG9udG5lZWRhLnB3OjUwMDAifQ"
            },
            "type": "public-key",
            "clientExtensionResults": {}
        }"""
        challenge = base64url_to_bytes(
            "1nhqyMkfGAQL-tTcsfpuoyq8hqeoHr0d9tDmjv1BuJ97YTA3FLWS5EdY4qUg-Mzruc2siBdyVlnFIPB1g0Hh2A"
        )
        rp_id = "dev2.dontneeda.pw"
        expected_origin = "https://dev2.dontneeda.pw:5000"

        verification = verify_registration_response(
            credential=credential,
            expected_challenge=challenge,
            expected_origin=expected_origin,
            expected_rp_id=rp_id,
        )

        assert verification.fmt == AttestationFormat.APPLE
        assert verification.credential_id == base64url_to_bytes("0yhsKG_gCzynIgNbvXWkqJKL8Uc")



from enum import Enum
from dataclasses import dataclass, field
from typing import List, Literal, Optional

from .cose import COSEAlgorithmIdentifier


################
#
# Fundamental data structures
#
################


class AuthenticatorTransport(str, Enum):
    """How an authenticator communicates to the client/browser.

    Members:
        `USB`: USB wired connection
        `NFC`: Near Field Communication
        `BLE`: Bluetooth Low Energy
        `INTERNAL`: Direct connection (read: a platform authenticator)
        `CABLE`: Cloud Assisted Bluetooth Low Energy
        `HYBRID`: A combination of (often separate) data-transport and proximity mechanisms

    https://www.w3.org/TR/webauthn-2/#enum-transport
    """

    USB = "usb"
    NFC = "nfc"
    BLE = "ble"
    SMART_CARD = "smart-card"
    INTERNAL = "internal"
    CABLE = "cable"
    HYBRID = "hybrid"


class AuthenticatorAttachment(str, Enum):
    """How an authenticator is connected to the client/browser.

    Members:
        `PLATFORM`: A non-removable authenticator, like TouchID or Windows Hello
        `CROSS_PLATFORM`: A "roaming" authenticator, like a YubiKey

    https://www.w3.org/TR/webauthn-2/#enumdef-authenticatorattachment
    """

    PLATFORM = "platform"
    CROSS_PLATFORM = "cross-platform"


class ResidentKeyRequirement(str, Enum):
    """The Relying Party's preference for the authenticator to create a dedicated "client-side" credential for it. Requiring an authenticator to store a dedicated credential should not be done lightly due to the limited storage capacity of some types of authenticators.

    Members:
        `DISCOURAGED`: The authenticator should not create a dedicated credential
        `PREFERRED`: The authenticator can create and store a dedicated credential, but if it doesn't that's alright too
        `REQUIRED`: The authenticator MUST create a dedicated credential. If it cannot, the RP is prepared for an error to occur.

    https://www.w3.org/TR/webauthn-2/#enum-residentKeyRequirement
    """

    DISCOURAGED = "discouraged"
    PREFERRED = "preferred"
    REQUIRED = "required"


class UserVerificationRequirement(str, Enum):
    """The degree to which the Relying Party wishes to verify a user's identity.

    Members:
        `REQUIRED`: User verification must occur
        `PREFERRED`: User verification would be great, but if not that's okay too
        `DISCOURAGED`: User verification should not occur, but it's okay if it does

    https://www.w3.org/TR/webauthn-2/#enumdef-userverificationrequirement
    """

    REQUIRED = "required"
    PREFERRED = "preferred"
    DISCOURAGED = "discouraged"


class AttestationConveyancePreference(str, Enum):
    """The Relying Party's interest in receiving an attestation statement.

    Members:
        `NONE`: The Relying Party isn't interested in receiving an attestation statement
        `INDIRECT`: The Relying Party is interested in an attestation statement, but the client is free to generate it as it sees fit
        `DIRECT`: The Relying Party is interested in an attestation statement generated directly by the authenticator
        `ENTERPRISE`: The Relying Party is interested in a statement with identifying information. Typically used within organizations

    https://www.w3.org/TR/webauthn-2/#enum-attestation-convey
    """

    NONE = "none"
    INDIRECT = "indirect"
    DIRECT = "direct"
    ENTERPRISE = "enterprise"


class PublicKeyCredentialType(str, Enum):
    """The type of credential that should be returned by an authenticator. There's but a single member because this is a specific subclass of a higher-level `CredentialType` that can be of other types.

    Members:
        `PUBLIC_KEY`: The literal string `"public-key"`

    https://www.w3.org/TR/webauthn-2/#enumdef-publickeycredentialtype
    """

    PUBLIC_KEY = "public-key"


class AttestationFormat(str, Enum):
    """The "syntax" of an attestation statement. Formats should be registered with the IANA and include documented signature verification steps.

    Members:
        `PACKED`
        `TPM`
        `ANDROID_KEY`
        `ANDROID_SAFETYNET`
        `FIDO_U2F`
        `APPLE`
        `NONE`

    https://www.iana.org/assignments/webauthn/webauthn.xhtml
    """

    PACKED = "packed"
    TPM = "tpm"
    ANDROID_KEY = "android-key"
    ANDROID_SAFETYNET = "android-safetynet"
    FIDO_U2F = "fido-u2f"
    APPLE = "apple"
    NONE = "none"


class ClientDataType(str, Enum):
    """Specific values included in authenticator registration and authentication responses to help avoid certain types of "signature confusion attacks".

    Members:
        `WEBAUTHN_CREATE`: The string "webauthn.create". Synonymous with `navigator.credentials.create()` in the browser
        `WEBAUTHN_GET`: The string "webauthn.get". Synonymous with `navigator.credentials.get()` in the browser

    https://www.w3.org/TR/webauthn-2/#dom-collectedclientdata-type
    """

    WEBAUTHN_CREATE = "webauthn.create"
    WEBAUTHN_GET = "webauthn.get"


class TokenBindingStatus(str, Enum):
    """
    https://www.w3.org/TR/webauthn-2/#dom-tokenbinding-status
    """

    PRESENT = "present"
    SUPPORTED = "supported"


class PublicKeyCredentialHint(str, Enum):
    """Categories of authenticators that Relying Parties can pass along to browsers during
    registration. Browsers that understand these values can optimize their modal experience to
    start the user off in a particular registration flow. These values are less strict than
    `authenticatorAttachment` (see `webauthn.helpers.strucAuthenticatorAttachment`)

    Members:
        `SECURITY_KEY`: A portable FIDO2 authenticator capable of being used on multiple devices via a USB or NFC connection
        `CLIENT_DEVICE`: The device that WebAuthn is being called on. Typically synonymous with platform authenticators
        `HYBRID`: A platform authenticator on a mobile device

    https://w3c.github.io/webauthn/#enumdef-publickeycredentialhint
    """

    SECURITY_KEY = "security-key"
    CLIENT_DEVICE = "client-device"
    HYBRID = "hybrid"


@dataclass
class TokenBinding:
    """
    https://www.w3.org/TR/webauthn-2/#dictdef-tokenbinding
    """

    status: TokenBindingStatus
    id: Optional[str] = None


@dataclass
class PublicKeyCredentialRpEntity:
    """Information about the Relying Party.

    Attributes:
        `name`: A user-readable name for the Relying Party
        (optional) `id`: A unique, constant value assigned to the Relying Party. Authenticators use this value to associate a credential with a particular Relying Party user

    https://www.w3.org/TR/webauthn-2/#dictdef-publickeycredentialrpentity
    """

    name: str
    id: Optional[str] = None


@dataclass
class PublicKeyCredentialUserEntity:
    """Information about a user of a Relying Party.

    Attributes:
        `id`: An "opaque byte sequence" that uniquely identifies a user. Typically something like a UUID, but never user-identifying like an email address. Cannot exceed 64 bytes. These bytes should be stored internally alongside your normal user identifier and only used for WebAuthn.
        `name`: A value which a user can see to determine which account this credential is associated with. A username or email address is fine here.
        `display_name`: A user-friendly representation of a user, like a full name.

    https://www.w3.org/TR/webauthn-2/#dictdef-publickeycredentialuserentity
    """

    id: bytes
    name: str
    display_name: str


@dataclass
class PublicKeyCredentialParameters:
    """Information about a cryptographic algorithm that may be used when creating a credential.

    Attributes:
        `type`: The literal string `"public-key"`
        `alg`: A numeric indicator of a particular algorithm

    https://www.w3.org/TR/webauthn-2/#dictdef-publickeycredentialparameters
    """

    type: Literal["public-key"]
    alg: COSEAlgorithmIdentifier


@dataclass
class PublicKeyCredentialDescriptor:
    """Information about a generated credential.

    Attributes:
        `type`: The literal string `"public-key"`
        `id`: The sequence of bytes representing the credential's ID
        (optional) `transports`: The types of connections to the client/browser the authenticator supports

    https://www.w3.org/TR/webauthn-2/#dictdef-publickeycredentialdescriptor
    """

    id: bytes
    type: Literal[PublicKeyCredentialType.PUBLIC_KEY] = PublicKeyCredentialType.PUBLIC_KEY
    transports: Optional[List[AuthenticatorTransport]] = None


@dataclass
class AuthenticatorSelectionCriteria:
    """A Relying Party's requirements for the types of authenticators that may interact with the client/browser.

    Attributes:
        (optional) `authenticator_attachment`: How the authenticator can be connected to the client/browser
        (optional) `resident_key`: Whether the authenticator should be able to store a credential on itself
        (optional) `require_resident_key`: DEPRECATED, set a value for `resident_key` instead
        (optional) `user_verification`: How the authenticator should be capable of determining user identity

    https://www.w3.org/TR/webauthn-2/#dictdef-authenticatorselectioncriteria
    """

    authenticator_attachment: Optional[AuthenticatorAttachment] = None
    resident_key: Optional[ResidentKeyRequirement] = None
    require_resident_key: Optional[bool] = False
    user_verification: Optional[UserVerificationRequirement] = (
        UserVerificationRequirement.PREFERRED
    )


@dataclass
class CollectedClientData:
    """Decoded ClientDataJSON

    Attributes:
        `type`: Either `"webauthn.create"` or `"webauthn.get"`, for registration and authentication ceremonies respectively
        `challenge`: The challenge passed to the authenticator within the options
        `origin`: The base domain with protocol on which the registration or authentication ceremony took place (e.g. "https://foo.bar")
        (optional) `cross_origin`: Whether or not the registration or authentication ceremony took place on a different origin (think within an <iframe>)
        (optional) `token_binding`: Information on the state of the Token Binding protocol

    https://www.w3.org/TR/webauthn-2/#dictdef-collectedclientdata
    """

    type: ClientDataType
    challenge: bytes
    origin: str
    cross_origin: Optional[bool] = None
    token_binding: Optional[TokenBinding] = None


################
#
# Registration
#
################


@dataclass
class PublicKeyCredentialCreationOptions:
    """Registration Options.

    Attributes:
        `rp`: Information about the Relying Party
        `user`: Information about the user
        `challenge`: A unique byte sequence to be returned by the authenticator. Helps prevent replay attacks
        `pub_key_cred_params`: Cryptographic algorithms supported by the Relying Party when verifying signatures
        (optional) `timeout`: How long the client/browser should give the user to interact with an authenticator
        (optional) `exclude_credentials`: A list of credentials associated with the user to prevent them from re-enrolling one of them
        (optional) `authenticator_selection`: Additional qualities about the authenticators the user can use to complete registration
        (optional) `hints`: Suggestions to the browser about the type of authenticator the user should try and register. Multiple values should be ordered by decreasing preference
        (optional) `attestation`: The Relying Party's desire for a declaration of an authenticator's provenance via attestation statement

    https://www.w3.org/TR/webauthn-2/#dictdef-publickeycredentialcreationoptions
    """

    rp: PublicKeyCredentialRpEntity
    user: PublicKeyCredentialUserEntity
    challenge: bytes
    pub_key_cred_params: List[PublicKeyCredentialParameters]
    timeout: Optional[int] = None
    exclude_credentials: Optional[List[PublicKeyCredentialDescriptor]] = None
    authenticator_selection: Optional[AuthenticatorSelectionCriteria] = None
    hints: Optional[List[PublicKeyCredentialHint]] = None
    attestation: AttestationConveyancePreference = AttestationConveyancePreference.NONE


@dataclass
class AuthenticatorAttestationResponse:
    """The `response` property on a registration credential.

    Attributes:
        `client_data_json`: Information the authenticator collects about the client/browser it communicates with
        `attestation_object`: Encoded information about an attestation
        (optional) `transports`: The authenticator's supported methods of communication with a client/browser

    https://www.w3.org/TR/webauthn-2/#authenticatorattestationresponse
    """

    client_data_json: bytes
    attestation_object: bytes
    # Optional in L2, but becomes required in L3. Play it safe until L3 becomes Recommendation
    transports: Optional[List[AuthenticatorTransport]] = None


@dataclass
class RegistrationCredential:
    """A registration-specific subclass of PublicKeyCredential returned from `navigator.credentials.create()`

    Attributes:
        `id`: The Base64URL-encoded representation of raw_id
        `raw_id`: A byte sequence representing the credential's unique identifier
        `response`: The authenticator's attestation data
        `type`: The literal string `"public-key"`

    https://www.w3.org/TR/webauthn-2/#publickeycredential
    """

    id: str
    raw_id: bytes
    response: AuthenticatorAttestationResponse
    authenticator_attachment: Optional[AuthenticatorAttachment] = None
    type: Literal[PublicKeyCredentialType.PUBLIC_KEY] = PublicKeyCredentialType.PUBLIC_KEY


@dataclass
class AttestationStatement:
    """A collection of all possible fields that may exist in an attestation statement. Combinations of these fields are specific to a particular attestation format.

    https://www.w3.org/TR/webauthn-2/#sctn-defined-attestation-formats

    TODO: Decide if this is acceptable, or if we want to split this up into multiple
    format-specific classes that define only the fields that are present for a given
    attestation format.
    """

    sig: Optional[bytes] = None
    x5c: Optional[List[bytes]] = None
    response: Optional[bytes] = None
    alg: Optional[COSEAlgorithmIdentifier] = None
    ver: Optional[str] = None
    cert_info: Optional[bytes] = None
    pub_area: Optional[bytes] = None


@dataclass
class AuthenticatorDataFlags:
    """Flags the authenticator will set about information contained within the `attestationObject.authData` property.

    Attributes:
        `up`: [U]ser was [P]resent
        `uv`: [U]ser was [V]erified
        `be`: [B]ackup [E]ligible
        `bs`: [B]ackup [S]tate
        `at`: [AT]tested credential is included
        `ed`: [E]xtension [D]ata is included

    https://www.w3.org/TR/webauthn-2/#flags
    """

    up: bool
    uv: bool
    be: bool
    bs: bool
    at: bool
    ed: bool


@dataclass
class AttestedCredentialData:
    """Information about a credential.

    Attributes:
        `aaguid`: A 128-bit identifier indicating the type and vendor of the authenticator
        `credential_id`: The ID of the private/public key pair generated by the authenticator
        `credential_public_key`: The public key generated by the authenticator

    https://www.w3.org/TR/webauthn-2/#attested-credential-data
    """

    aaguid: bytes
    credential_id: bytes
    credential_public_key: bytes


@dataclass
class AuthenticatorData:
    """Context the authenticator provides about itself and the environment in which the registration or authentication ceremony took place.

    Attributes:
        `rp_id_hash`: A SHA-256 hash of the website origin on which the registration or authentication ceremony took place
        `flags`: Properties about the user and registration, where applicable
        `sign_count`: The number of times the credential was used
        (optional) `attested_credential_data`: Information about the credential created during a registration ceremony
        (optional) `extensions`: CBOR-encoded extension data corresponding to extensions specified in the registration or authentication ceremony options

    https://www.w3.org/TR/webauthn-2/#sctn-attestation
    https://www.w3.org/TR/webauthn-2/#sctn-attested-credential-data
    """

    rp_id_hash: bytes
    flags: AuthenticatorDataFlags
    sign_count: int
    attested_credential_data: Optional[AttestedCredentialData] = None
    extensions: Optional[bytes] = None


@dataclass
class AttestationObject:
    """Information about an attestation, including a statement and authenticator data.

    Attributes:
        `fmt`: The attestation statement's format
        `att_stmt`: An attestation statement to be verified according to the format
        `auth_data`: Contextual information provided by authenticator

    https://www.w3.org/TR/webauthn-2/#sctn-attestation
    """

    fmt: AttestationFormat
    auth_data: AuthenticatorData
    att_stmt: AttestationStatement = field(default_factory=AttestationStatement)


################
#
# Authentication
#
################


@dataclass
class PublicKeyCredentialRequestOptions:
    """Authentication Options.

    Attributes:
        `challenge`: A unique byte sequence to be returned by the authenticator. Helps prevent replay attacks
        (optional) `timeout`: How long the client/browser should give the user to interact with an authenticator
        (optional) `rp_id`: The unique, constant identifier assigned to the Relying Party
        (optional) `allow_credentials`: A list of credentials associated with the user that they can use to complete the authentication
        (optional) `user_verification`: How the authenticator should be capable of determining user identity

    https://www.w3.org/TR/webauthn-2/#dictionary-assertion-options
    """

    challenge: bytes
    timeout: Optional[int] = None
    rp_id: Optional[str] = None
    allow_credentials: Optional[List[PublicKeyCredentialDescriptor]] = None
    user_verification: Optional[UserVerificationRequirement] = (
        UserVerificationRequirement.PREFERRED
    )


@dataclass
class AuthenticatorAssertionResponse:
    """The `response` property on an authentication credential.

    Attributes:
        `client_data_json`: Information the authenticator collects about the client/browser it communicates with
        `authenticator_data`: Contextual information provided by authenticator
        `signature`: A byte sequence signed by the authenticator's private key, to be verified with a user's public key
        (optional) `user_handle`: The user ID specified for the user during attestation

    https://www.w3.org/TR/webauthn-2/#authenticatorassertionresponse
    """

    client_data_json: bytes
    authenticator_data: bytes
    signature: bytes
    user_handle: Optional[bytes] = None


@dataclass
class AuthenticationCredential:
    """An authentication-specific subclass of PublicKeyCredential. Returned from `navigator.credentials.get()`

    Attributes:
        `id`: The Base64URL-encoded representation of raw_id
        `raw_id`: A byte sequence representing the credential's unique identifier
        `response`: The authenticator's assertion data
        `type`: The literal string `"public-key"`

    https://www.w3.org/TR/webauthn-2/#publickeycredential
    """

    id: str
    raw_id: bytes
    response: AuthenticatorAssertionResponse
    authenticator_attachment: Optional[AuthenticatorAttachment] = None
    type: Literal[PublicKeyCredentialType.PUBLIC_KEY] = PublicKeyCredentialType.PUBLIC_KEY


################
#
# Credential Backup State
#
################


class CredentialDeviceType(str, Enum):
    """A determination of the number of devices a credential can be used from

    Members:
        `SINGLE_DEVICE`: A credential that is bound to a single device
        `MULTI_DEVICE`: A credential that can be used from multiple devices (e.g. passkeys)

    https://w3c.github.io/webauthn/#sctn-credential-backup (L3 Draft)
    """

    SINGLE_DEVICE = "single_device"
    MULTI_DEVICE = "multi_device"



from unittest import TestCase

from webauthn import verify_authentication_response
from webauthn.helpers import base64url_to_bytes, parse_authentication_credential_json
from webauthn.helpers.exceptions import InvalidAuthenticationResponse


class TestVerifyAuthenticationResponse(TestCase):
    def test_verify_authentication_response_with_EC2_public_key(self):
        credential = """{
            "id": "EDx9FfAbp4obx6oll2oC4-CZuDidRVV4gZhxC529ytlnqHyqCStDUwfNdm1SNHAe3X5KvueWQdAX3x9R1a2b9Q",
            "rawId": "EDx9FfAbp4obx6oll2oC4-CZuDidRVV4gZhxC529ytlnqHyqCStDUwfNdm1SNHAe3X5KvueWQdAX3x9R1a2b9Q",
            "response": {
                "authenticatorData": "SZYN5YgOjGh0NBcPZHZgW4_krrmihjLHmVzzuoMdl2MBAAAATg",
                "clientDataJSON": "eyJjaGFsbGVuZ2UiOiJ4aTMwR1BHQUZZUnhWRHBZMXNNMTBEYUx6VlFHNjZudi1fN1JVYXpIMHZJMll2RzhMWWdERW52TjVmWlpOVnV2RUR1TWk5dGUzVkxxYjQyTjBma0xHQSIsImNsaWVudEV4dGVuc2lvbnMiOnt9LCJoYXNoQWxnb3JpdGhtIjoiU0hBLTI1NiIsIm9yaWdpbiI6Imh0dHA6Ly9sb2NhbGhvc3Q6NTAwMCIsInR5cGUiOiJ3ZWJhdXRobi5nZXQifQ",
                "signature": "MEUCIGisVZOBapCWbnJJvjelIzwpixxIwkjCCb5aCHafQu68AiEA88v-2pJNNApPFwAKFiNuf82-2hBxYW5kGwVweeoxCwo"
            },
            "type": "public-key",
            "clientExtensionResults": {}
        }"""
        challenge = base64url_to_bytes(
            "xi30GPGAFYRxVDpY1sM10DaLzVQG66nv-_7RUazH0vI2YvG8LYgDEnvN5fZZNVuvEDuMi9te3VLqb42N0fkLGA"
        )
        expected_rp_id = "localhost"
        expected_origin = "http://localhost:5000"
        credential_public_key = base64url_to_bytes(
            "pQECAyYgASFYIIeDTe-gN8A-zQclHoRnGFWN8ehM1b7yAsa8I8KIvmplIlgg4nFGT5px8o6gpPZZhO01wdy9crDSA_Ngtkx0vGpvPHI"
        )
        sign_count = 77

        verification = verify_authentication_response(
            credential=credential,
            expected_challenge=challenge,
            expected_rp_id=expected_rp_id,
            expected_origin=expected_origin,
            credential_public_key=credential_public_key,
            credential_current_sign_count=sign_count,
        )

        assert verification.credential_id == base64url_to_bytes(
            "EDx9FfAbp4obx6oll2oC4-CZuDidRVV4gZhxC529ytlnqHyqCStDUwfNdm1SNHAe3X5KvueWQdAX3x9R1a2b9Q"
        )
        assert verification.new_sign_count == 78
        assert verification.credential_backed_up == False
        assert verification.credential_device_type == "single_device"
        assert not verification.user_verified

    def test_verify_authentication_response_with_RSA_public_key(self):
        credential = """{
            "id": "ZoIKP1JQvKdrYj1bTUPJ2eTUsbLeFkv-X5xJQNr4k6s",
            "rawId": "ZoIKP1JQvKdrYj1bTUPJ2eTUsbLeFkv-X5xJQNr4k6s",
            "response": {
                "authenticatorData": "SZYN5YgOjGh0NBcPZHZgW4_krrmihjLHmVzzuoMdl2MFAAAAAQ",
                "clientDataJSON": "eyJ0eXBlIjoid2ViYXV0aG4uZ2V0IiwiY2hhbGxlbmdlIjoiaVBtQWkxUHAxWEw2b0FncTNQV1p0WlBuWmExekZVRG9HYmFRMF9LdlZHMWxGMnMzUnRfM280dVN6Y2N5MHRtY1RJcFRUVDRCVTFULUk0bWFhdm5kalEiLCJvcmlnaW4iOiJodHRwOi8vbG9jYWxob3N0OjUwMDAiLCJjcm9zc09yaWdpbiI6ZmFsc2V9",
                "signature": "iOHKX3erU5_OYP_r_9HLZ-CexCE4bQRrxM8WmuoKTDdhAnZSeTP0sjECjvjfeS8MJzN1ArmvV0H0C3yy_FdRFfcpUPZzdZ7bBcmPh1XPdxRwY747OrIzcTLTFQUPdn1U-izCZtP_78VGw9pCpdMsv4CUzZdJbEcRtQuRS03qUjqDaovoJhOqEBmxJn9Wu8tBi_Qx7A33RbYjlfyLm_EDqimzDZhyietyop6XUcpKarKqVH0M6mMrM5zTjp8xf3W7odFCadXEJg-ERZqFM0-9Uup6kJNLbr6C5J4NDYmSm3HCSA6lp2iEiMPKU8Ii7QZ61kybXLxsX4w4Dm3fOLjmDw",
                "userHandle": "T1RWa1l6VXdPRFV0WW1NNVlTMDBOVEkxTFRnd056Z3RabVZpWVdZNFpEVm1ZMk5p"
            },
            "type": "public-key",
            "clientExtensionResults": {}
        }"""
        challenge = base64url_to_bytes(
            "iPmAi1Pp1XL6oAgq3PWZtZPnZa1zFUDoGbaQ0_KvVG1lF2s3Rt_3o4uSzccy0tmcTIpTTT4BU1T-I4maavndjQ"
        )
        expected_rp_id = "localhost"
        expected_origin = "http://localhost:5000"
        credential_public_key = base64url_to_bytes(
            "pAEDAzkBACBZAQDfV20epzvQP-HtcdDpX-cGzdOxy73WQEvsU7Dnr9UWJophEfpngouvgnRLXaEUn_d8HGkp_HIx8rrpkx4BVs6X_B6ZjhLlezjIdJbLbVeb92BaEsmNn1HW2N9Xj2QM8cH-yx28_vCjf82ahQ9gyAr552Bn96G22n8jqFRQKdVpO-f-bvpvaP3IQ9F5LCX7CUaxptgbog1SFO6FI6ob5SlVVB00lVXsaYg8cIDZxCkkENkGiFPgwEaZ7995SCbiyCpUJbMqToLMgojPkAhWeyktu7TlK6UBWdJMHc3FPAIs0lH_2_2hKS-mGI1uZAFVAfW1X-mzKL0czUm2P1UlUox7IUMBAAE"
        )
        sign_count = 0

        verification = verify_authentication_response(
            credential=credential,
            expected_challenge=challenge,
            expected_rp_id=expected_rp_id,
            expected_origin=expected_origin,
            credential_public_key=credential_public_key,
            credential_current_sign_count=sign_count,
            require_user_verification=True,
        )

        assert verification.new_sign_count == 1
        assert verification.user_verified

    def test_raises_exception_on_incorrect_public_key(self):
        credential = """{
            "id": "FviUBZA3FGMxEm3A1K2T8MhuEBLp4qQsV9ScAKYrpdw2kbGnqx24tF4ev6PEHEYC3g8z6HMJh7dYHe3Uuq7_8Q",
            "rawId": "FviUBZA3FGMxEm3A1K2T8MhuEBLp4qQsV9ScAKYrpdw2kbGnqx24tF4ev6PEHEYC3g8z6HMJh7dYHe3Uuq7_8Q",
            "response": {
                "authenticatorData": "SZYN5YgOjGh0NBcPZHZgW4_krrmihjLHmVzzuoMdl2MFAAAAJA",
                "clientDataJSON": "eyJ0eXBlIjoid2ViYXV0aG4uZ2V0IiwiY2hhbGxlbmdlIjoienNmaU1aajE2VFVWQ3JUNXREUllYZFlsVXJKcDd6bl9VTmQ1Tm1Cb2NQYzRJMmRLWmJlRVdwd0JBd0E0czZvSGtWWDZfbHlfamdwNzQzZHlpV0hZWXciLCJvcmlnaW4iOiJodHRwOi8vbG9jYWxob3N0OjUwMDAiLCJjcm9zc09yaWdpbiI6ZmFsc2V9",
                "signature": "MEQCIBX9B1LaLaQ0LYJsRv7cOyMS-Do1rJfFJoF9oO1tHMA4AiBRKdNneMKPlN53i8uoTZ5y9Gj4ORZySmiercS38655_g"
            },
            "type": "public-key",
            "clientExtensionResults": {}
        }"""
        challenge = base64url_to_bytes(
            "zsfiMZj16TUVCrT5tDRYXdYlUrJp7zn_UNd5NmBocPc4I2dKZbeEWpwBAwA4s6oHkVX6_ly_jgp743dyiWHYYw"
        )
        expected_rp_id = "localhost"
        expected_origin = "http://localhost:5000"
        credential_public_key = base64url_to_bytes(
            "pAEDAzkBACBZAQDfV20epzvQP-HtcdDpX-cGzdOxy73WQEvsU7Dnr9UWJophEfpngouvgnRLXaEUn_d8HGkp_HIx8rrpkx4BVs6X_B6ZjhLlezjIdJbLbVeb92BaEsmNn1HW2N9Xj2QM8cH-yx28_vCjf82ahQ9gyAr552Bn96G22n8jqFRQKdVpO-f-bvpvaP3IQ9F5LCX7CUaxptgbog1SFO6FI6ob5SlVVB00lVXsaYg8cIDZxCkkENkGiFPgwEaZ7995SCbiyCpUJbMqToLMgojPkAhWeyktu7TlK6UBWdJMHc3FPAIs0lH_2_2hKS-mGI1uZAFVAfW1X-mzKL0czUm2P1UlUox7IUMBAAE"
        )
        sign_count = 35

        with self.assertRaisesRegex(
            InvalidAuthenticationResponse,
            "Could not verify authentication signature",
        ):
            verify_authentication_response(
                credential=credential,
                expected_challenge=challenge,
                expected_rp_id=expected_rp_id,
                expected_origin=expected_origin,
                credential_public_key=credential_public_key,
                credential_current_sign_count=sign_count,
                require_user_verification=True,
            )

    def test_raises_exception_on_uv_required_but_false(self):
        credential = """{
            "id": "4-5MZF69j3n2B6Z99dUN0fNrAQmrjELJIebWVw8aKfw1EQKg28Tx40R_kw-1pcrfSgJFKm3mCtAtBgSRWgDMng",
            "rawId": "4-5MZF69j3n2B6Z99dUN0fNrAQmrjELJIebWVw8aKfw1EQKg28Tx40R_kw-1pcrfSgJFKm3mCtAtBgSRWgDMng",
            "response": {
                "authenticatorData": "SZYN5YgOjGh0NBcPZHZgW4_krrmihjLHmVzzuoMdl2MBAAAAIQ",
                "clientDataJSON": "eyJ0eXBlIjoid2ViYXV0aG4uZ2V0IiwiY2hhbGxlbmdlIjoidW1HZW1YSklQQlhQeGtEOEhqYW51djlCRG9yOFo3TzNhUGR0T2dNQ2RXNFBBZnFEWDQzRUZsaHJzRjBQVzkwZGY1enJnYnQ3WVZNUkFhMjd0Q2RIenciLCJvcmlnaW4iOiJodHRwOi8vbG9jYWxob3N0OjUwMDAiLCJjcm9zc09yaWdpbiI6ZmFsc2V9",
                "signature": "MEUCIGp5ADnU_SFvT4J_bKvQJ4Pc1GmANhbYq5GioOLjyUrxAiEA6Kk5qAZb8MLY-jyTiJLr_R9Fke02UHkxsRB0dnZt2X8"
            },
            "type": "public-key",
            "clientExtensionResults": {}
        }"""
        challenge = base64url_to_bytes(
            "umGemXJIPBXPxkD8Hjanuv9BDor8Z7O3aPdtOgMCdW4PAfqDX43EFlhrsF0PW90df5zrgbt7YVMRAa27tCdHzw"
        )
        expected_rp_id = "localhost"
        expected_origin = "http://localhost:5000"
        credential_public_key = base64url_to_bytes(
            "pQECAyYgASFYIOQ5TKpXJR2cV76Wgfge9BkLkEhLxVjhFjM1jKHYOcqpIlggaiNy1blt3OU8Hsmg041HUYP7eajgL7fk3nSuTEjYCwU"
        )
        sign_count = 32

        with self.assertRaisesRegex(
            InvalidAuthenticationResponse,
            "User verification is required but user was not verified",
        ):
            verify_authentication_response(
                credential=credential,
                expected_challenge=challenge,
                expected_rp_id=expected_rp_id,
                expected_origin=expected_origin,
                credential_public_key=credential_public_key,
                credential_current_sign_count=sign_count,
                require_user_verification=True,
            )

    def test_verify_authentication_response_with_OKP_public_key(self):
        credential = """{
            "id": "fq9Nj0nS24B5y6Pkw_h3-9GEAEA3-0LBPxE2zvTdLjDqtSeCSNYFe9VMRueSpAZxT3YDc6L1lWXdQNwI-sVNYrefEcRR1Nsb_0jpHE955WEtFud2xxZg3MvoLMxHLet63i5tajd1fHtP7I-00D6cehM8ZWlLp2T3s9lfZgVIFcA",
            "rawId": "fq9Nj0nS24B5y6Pkw_h3-9GEAEA3-0LBPxE2zvTdLjDqtSeCSNYFe9VMRueSpAZxT3YDc6L1lWXdQNwI-sVNYrefEcRR1Nsb_0jpHE955WEtFud2xxZg3MvoLMxHLet63i5tajd1fHtP7I-00D6cehM8ZWlLp2T3s9lfZgVIFcA",
            "response": {
                "authenticatorData": "SZYN5YgOjGh0NBcPZHZgW4_krrmihjLHmVzzuoMdl2MBAAAABw",
                "clientDataJSON": "eyJ0eXBlIjoid2ViYXV0aG4uZ2V0IiwiY2hhbGxlbmdlIjoiZVo0ZWVBM080ank1Rkl6cURhU0o2SkROR3UwYkJjNXpJMURqUV9rTHNvMVdOcWtHNms1bUNZZjFkdFFoVlVpQldaV2xaa3pSNU1GZWVXQ3BKUlVOWHciLCJvcmlnaW4iOiJodHRwOi8vbG9jYWxob3N0OjUwMDAiLCJjcm9zc09yaWdpbiI6ZmFsc2V9",
                "signature": "RRWV8mYDRvK7YdQgdtZD4pJ2dh1D_IWZ_D6jsZo6FHJBoenbj0CVT5nA20vUzlRhN4R6dOEUHmUwP1F8eRBhBg"
            },
            "type": "public-key",
            "clientExtensionResults": {}
        }"""
        challenge = base64url_to_bytes(
            "eZ4eeA3O4jy5FIzqDaSJ6JDNGu0bBc5zI1DjQ_kLso1WNqkG6k5mCYf1dtQhVUiBWZWlZkzR5MFeeWCpJRUNXw"
        )
        expected_rp_id = "localhost"
        expected_origin = "http://localhost:5000"
        credential_public_key = base64url_to_bytes(
            "pAEBAycgBiFYIMz6_SUFLiDid2Yhlq0YboyJ-CDrIrNpkPUGmJp4D3Dp"
        )
        sign_count = 3

        verification = verify_authentication_response(
            credential=credential,
            expected_challenge=challenge,
            expected_rp_id=expected_rp_id,
            expected_origin=expected_origin,
            credential_public_key=credential_public_key,
            credential_current_sign_count=sign_count,
        )

        assert verification.new_sign_count == 7

    def test_supports_multiple_expected_origins(self) -> None:
        credential = """{
            "id": "AXmOjWWZH67pgl5_gAbKVBqoL2dyHHGEWZLspIsCwULG0hZ3HyuGgvkaRcSOLq9W72XtegcvFYXIdlafrilbtVnx2Q14gNbfSQQP2sgNEAif4MjHtGpeVB0BfFawCs85Y3XY_j4sxthVnyTY_Q",
            "rawId": "AXmOjWWZH67pgl5_gAbKVBqoL2dyHHGEWZLspIsCwULG0hZ3HyuGgvkaRcSOLq9W72XtegcvFYXIdlafrilbtVnx2Q14gNbfSQQP2sgNEAif4MjHtGpeVB0BfFawCs85Y3XY_j4sxthVnyTY_Q",
            "response": {
                "authenticatorData": "SZYN5YgOjGh0NBcPZHZgW4_krrmihjLHmVzzuoMdl2MFYN-Mog",
                "clientDataJSON": "eyJ0eXBlIjoid2ViYXV0aG4uZ2V0IiwiY2hhbGxlbmdlIjoiNnpyU1JYOEN4d1BTWEVBclh5WEwydHBiNnJCN1N0YXIwckxWSWo1cnZmNzRtWktGNWlyNzE1WG1nejV0QV9HeUhleE40b1hmclE4ODlBclZDTGFSZEEiLCJvcmlnaW4iOiJodHRwOi8vbG9jYWxob3N0OjUwMDAiLCJjcm9zc09yaWdpbiI6ZmFsc2V9",
                "signature": "MEUCIQDBqeI274exaKWGQz37g7yo1--TVcZSCcYVftZ1AnEJkQIgNw-nlx-_U9rVfFfER8oX6BlYZTuPFyGaL_wCDY23s0E",
                "userHandle": "TldNMFlqYzNOVFF0WW1NNE5DMDBaakprTFRrME9EVXROR05rTnpreVkyTTROVEUz"
            },
            "type": "public-key",
            "clientExtensionResults": {}
        }"""

        challenge = base64url_to_bytes(
            "6zrSRX8CxwPSXEArXyXL2tpb6rB7Star0rLVIj5rvf74mZKF5ir715Xmgz5tA_GyHexN4oXfrQ889ArVCLaRdA"
        )
        expected_rp_id = "localhost"
        expected_origin = ["https://foo.bar", "http://localhost:5000"]
        credential_public_key = base64url_to_bytes(
            "pQECAyYgASFYIFm1Py-FzzFOuwXbRbTr95SiDxuB1BkZsEEJxFhquzqkIlggL1U1T713Jo_2muzhXvpbwRNdoAs8CYK6PflvY1MBdCI"
        )
        sign_count = 1625263263

        verification = verify_authentication_response(
            credential=credential,
            expected_challenge=challenge,
            expected_rp_id=expected_rp_id,
            expected_origin=expected_origin,
            credential_public_key=credential_public_key,
            credential_current_sign_count=sign_count,
        )

        assert verification.credential_id == base64url_to_bytes(
            "AXmOjWWZH67pgl5_gAbKVBqoL2dyHHGEWZLspIsCwULG0hZ3HyuGgvkaRcSOLq9W72XtegcvFYXIdlafrilbtVnx2Q14gNbfSQQP2sgNEAif4MjHtGpeVB0BfFawCs85Y3XY_j4sxthVnyTY_Q"
        )

    def test_supports_already_parsed_credential(self) -> None:
        parsed_credential = parse_authentication_credential_json(
            """{
            "id": "ZoIKP1JQvKdrYj1bTUPJ2eTUsbLeFkv-X5xJQNr4k6s",
            "rawId": "ZoIKP1JQvKdrYj1bTUPJ2eTUsbLeFkv-X5xJQNr4k6s",
            "response": {
                "authenticatorData": "SZYN5YgOjGh0NBcPZHZgW4_krrmihjLHmVzzuoMdl2MFAAAAAQ",
                "clientDataJSON": "eyJ0eXBlIjoid2ViYXV0aG4uZ2V0IiwiY2hhbGxlbmdlIjoiaVBtQWkxUHAxWEw2b0FncTNQV1p0WlBuWmExekZVRG9HYmFRMF9LdlZHMWxGMnMzUnRfM280dVN6Y2N5MHRtY1RJcFRUVDRCVTFULUk0bWFhdm5kalEiLCJvcmlnaW4iOiJodHRwOi8vbG9jYWxob3N0OjUwMDAiLCJjcm9zc09yaWdpbiI6ZmFsc2V9",
                "signature": "iOHKX3erU5_OYP_r_9HLZ-CexCE4bQRrxM8WmuoKTDdhAnZSeTP0sjECjvjfeS8MJzN1ArmvV0H0C3yy_FdRFfcpUPZzdZ7bBcmPh1XPdxRwY747OrIzcTLTFQUPdn1U-izCZtP_78VGw9pCpdMsv4CUzZdJbEcRtQuRS03qUjqDaovoJhOqEBmxJn9Wu8tBi_Qx7A33RbYjlfyLm_EDqimzDZhyietyop6XUcpKarKqVH0M6mMrM5zTjp8xf3W7odFCadXEJg-ERZqFM0-9Uup6kJNLbr6C5J4NDYmSm3HCSA6lp2iEiMPKU8Ii7QZ61kybXLxsX4w4Dm3fOLjmDw",
                "userHandle": "T1RWa1l6VXdPRFV0WW1NNVlTMDBOVEkxTFRnd056Z3RabVZpWVdZNFpEVm1ZMk5p"
            },
            "type": "public-key",
            "clientExtensionResults": {}
        }"""
        )
        challenge = base64url_to_bytes(
            "iPmAi1Pp1XL6oAgq3PWZtZPnZa1zFUDoGbaQ0_KvVG1lF2s3Rt_3o4uSzccy0tmcTIpTTT4BU1T-I4maavndjQ"
        )
        expected_rp_id = "localhost"
        expected_origin = "http://localhost:5000"
        credential_public_key = base64url_to_bytes(
            "pAEDAzkBACBZAQDfV20epzvQP-HtcdDpX-cGzdOxy73WQEvsU7Dnr9UWJophEfpngouvgnRLXaEUn_d8HGkp_HIx8rrpkx4BVs6X_B6ZjhLlezjIdJbLbVeb92BaEsmNn1HW2N9Xj2QM8cH-yx28_vCjf82ahQ9gyAr552Bn96G22n8jqFRQKdVpO-f-bvpvaP3IQ9F5LCX7CUaxptgbog1SFO6FI6ob5SlVVB00lVXsaYg8cIDZxCkkENkGiFPgwEaZ7995SCbiyCpUJbMqToLMgojPkAhWeyktu7TlK6UBWdJMHc3FPAIs0lH_2_2hKS-mGI1uZAFVAfW1X-mzKL0czUm2P1UlUox7IUMBAAE"
        )
        sign_count = 0

        verification = verify_authentication_response(
            credential=parsed_credential,
            expected_challenge=challenge,
            expected_rp_id=expected_rp_id,
            expected_origin=expected_origin,
            credential_public_key=credential_public_key,
            credential_current_sign_count=sign_count,
            require_user_verification=True,
        )

        assert verification.new_sign_count == 1

    def test_supports_dict_credential(self) -> None:
        credential = {
            "id": "ZoIKP1JQvKdrYj1bTUPJ2eTUsbLeFkv-X5xJQNr4k6s",
            "rawId": "ZoIKP1JQvKdrYj1bTUPJ2eTUsbLeFkv-X5xJQNr4k6s",
            "response": {
                "authenticatorData": "SZYN5YgOjGh0NBcPZHZgW4_krrmihjLHmVzzuoMdl2MFAAAAAQ",
                "clientDataJSON": "eyJ0eXBlIjoid2ViYXV0aG4uZ2V0IiwiY2hhbGxlbmdlIjoiaVBtQWkxUHAxWEw2b0FncTNQV1p0WlBuWmExekZVRG9HYmFRMF9LdlZHMWxGMnMzUnRfM280dVN6Y2N5MHRtY1RJcFRUVDRCVTFULUk0bWFhdm5kalEiLCJvcmlnaW4iOiJodHRwOi8vbG9jYWxob3N0OjUwMDAiLCJjcm9zc09yaWdpbiI6ZmFsc2V9",
                "signature": "iOHKX3erU5_OYP_r_9HLZ-CexCE4bQRrxM8WmuoKTDdhAnZSeTP0sjECjvjfeS8MJzN1ArmvV0H0C3yy_FdRFfcpUPZzdZ7bBcmPh1XPdxRwY747OrIzcTLTFQUPdn1U-izCZtP_78VGw9pCpdMsv4CUzZdJbEcRtQuRS03qUjqDaovoJhOqEBmxJn9Wu8tBi_Qx7A33RbYjlfyLm_EDqimzDZhyietyop6XUcpKarKqVH0M6mMrM5zTjp8xf3W7odFCadXEJg-ERZqFM0-9Uup6kJNLbr6C5J4NDYmSm3HCSA6lp2iEiMPKU8Ii7QZ61kybXLxsX4w4Dm3fOLjmDw",
                "userHandle": "T1RWa1l6VXdPRFV0WW1NNVlTMDBOVEkxTFRnd056Z3RabVZpWVdZNFpEVm1ZMk5p",
            },
            "type": "public-key",
            "clientExtensionResults": {},
        }
        challenge = base64url_to_bytes(
            "iPmAi1Pp1XL6oAgq3PWZtZPnZa1zFUDoGbaQ0_KvVG1lF2s3Rt_3o4uSzccy0tmcTIpTTT4BU1T-I4maavndjQ"
        )
        expected_rp_id = "localhost"
        expected_origin = "http://localhost:5000"
        credential_public_key = base64url_to_bytes(
            "pAEDAzkBACBZAQDfV20epzvQP-HtcdDpX-cGzdOxy73WQEvsU7Dnr9UWJophEfpngouvgnRLXaEUn_d8HGkp_HIx8rrpkx4BVs6X_B6ZjhLlezjIdJbLbVeb92BaEsmNn1HW2N9Xj2QM8cH-yx28_vCjf82ahQ9gyAr552Bn96G22n8jqFRQKdVpO-f-bvpvaP3IQ9F5LCX7CUaxptgbog1SFO6FI6ob5SlVVB00lVXsaYg8cIDZxCkkENkGiFPgwEaZ7995SCbiyCpUJbMqToLMgojPkAhWeyktu7TlK6UBWdJMHc3FPAIs0lH_2_2hKS-mGI1uZAFVAfW1X-mzKL0czUm2P1UlUox7IUMBAAE"
        )
        sign_count = 0

        verification = verify_authentication_response(
            credential=credential,
            expected_challenge=challenge,
            expected_rp_id=expected_rp_id,
            expected_origin=expected_origin,
            credential_public_key=credential_public_key,
            credential_current_sign_count=sign_count,
            require_user_verification=True,
        )

        assert verification.new_sign_count == 1



from unittest import TestCase

from webauthn.helpers import base64url_to_bytes
from webauthn.helpers.aaguid_to_string import aaguid_to_string


class TestAAGUIDToString(TestCase):
    def test_converts_bytes_to_uuid_format(self):
        aaguid_bytes = base64url_to_bytes("AAAAAAAAAAAAAAAAAAAAAA")

        output = aaguid_to_string(aaguid_bytes)

        assert output == "00000000-0000-0000-0000-000000000000"



from .map_tpm_manufacturer import map_tpm_manufacturer_id
from .parse_cert_info import parse_cert_info
from .parse_pub_area import parse_pub_area

__all__ = ["map_tpm_manufacturer_id", "parse_cert_info", "parse_pub_area"]



from typing import Any

import cbor2

from .exceptions import InvalidCBORData


def parse_cbor(data: bytes) -> Any:
    """
    Attempt to decode CBOR-encoded data.

    Raises:
        `helpers.exceptions.InvalidCBORData` if data cannot be decoded
    """
    try:
        to_return = cbor2.loads(data)
    except Exception as exc:
        raise InvalidCBORData("Could not decode CBOR data") from exc

    return to_return



class WebAuthnException(Exception):
    pass


class InvalidRegistrationOptions(WebAuthnException):
    pass


class InvalidRegistrationResponse(WebAuthnException):
    pass


class InvalidAuthenticationOptions(WebAuthnException):
    pass


class InvalidAuthenticationResponse(WebAuthnException):
    pass


class InvalidPublicKeyStructure(WebAuthnException):
    pass


class UnsupportedPublicKeyType(WebAuthnException):
    pass


class InvalidJSONStructure(WebAuthnException):
    pass


class InvalidAuthenticatorDataStructure(WebAuthnException):
    pass


class SignatureVerificationException(WebAuthnException):
    pass


class UnsupportedAlgorithm(WebAuthnException):
    pass


class UnsupportedPublicKey(WebAuthnException):
    pass


class UnsupportedEC2Curve(WebAuthnException):
    pass


class InvalidTPMPubAreaStructure(WebAuthnException):
    pass


class InvalidTPMCertInfoStructure(WebAuthnException):
    pass


class InvalidCertificateChain(WebAuthnException):
    pass


class InvalidBackupFlags(WebAuthnException):
    pass


class InvalidCBORData(WebAuthnException):
    pass



from unittest import TestCase

from webauthn.helpers import base64url_to_bytes, options_to_json
from webauthn.helpers.exceptions import InvalidJSONStructure
from webauthn.helpers.structs import (
    AuthenticatorTransport,
    AuthenticatorAttachment,
    AttestationConveyancePreference,
    AuthenticatorSelectionCriteria,
    PublicKeyCredentialDescriptor,
    ResidentKeyRequirement,
    PublicKeyCredentialHint,
    PublicKeyCredentialRpEntity,
    PublicKeyCredentialUserEntity,
    UserVerificationRequirement,
    PublicKeyCredentialParameters,
)
from webauthn.helpers.cose import COSEAlgorithmIdentifier
from webauthn.helpers.parse_registration_options_json import parse_registration_options_json
from webauthn.registration.generate_registration_options import generate_registration_options


class TestParseRegistrationOptionsJSON(TestCase):
    maxDiff = None

    def test_returns_parsed_options_simple(self) -> None:
        parsed = parse_registration_options_json(
            {
                "rp": {"name": "Example Co", "id": "example.com"},
                "user": {
                    "id": "vEC5nFXSxpc_W68bX59JeD3c_-1XDJ5RblcWjY3Tx7RvfC0rkB19UWadf6wDEWG8T1ztksOYMim0sJIn6z_5tw",
                    "name": "bob",
                    "displayName": "bob",
                },
                "challenge": "scb_z5GweYijAT2ppsB0HAklsw96fPs_tOWh-myqkOeb9rcvhWBwUZ56J3t3eocgjHkS4Mf3XeXTOQc1ySvk5w",
                "pubKeyCredParams": [{"type": "public-key", "alg": -36}],
                "timeout": 60000,
                "excludeCredentials": [],
                "attestation": "none",
            }
        )

        self.assertEqual(
            parsed.rp, PublicKeyCredentialRpEntity(id="example.com", name="Example Co")
        )
        self.assertEqual(
            parsed.user,
            PublicKeyCredentialUserEntity(
                id=base64url_to_bytes(
                    "vEC5nFXSxpc_W68bX59JeD3c_-1XDJ5RblcWjY3Tx7RvfC0rkB19UWadf6wDEWG8T1ztksOYMim0sJIn6z_5tw"
                ),
                name="bob",
                display_name="bob",
            ),
        )
        self.assertEqual(parsed.attestation, AttestationConveyancePreference.NONE)
        self.assertEqual(parsed.authenticator_selection, None)
        self.assertEqual(
            parsed.challenge,
            base64url_to_bytes(
                "scb_z5GweYijAT2ppsB0HAklsw96fPs_tOWh-myqkOeb9rcvhWBwUZ56J3t3eocgjHkS4Mf3XeXTOQc1ySvk5w"
            ),
        )
        self.assertEqual(parsed.exclude_credentials, [])
        self.assertEqual(
            parsed.pub_key_cred_params,
            [
                PublicKeyCredentialParameters(
                    alg=COSEAlgorithmIdentifier.ECDSA_SHA_512,
                    type="public-key",
                )
            ],
        )
        self.assertEqual(parsed.timeout, 60000)

    def test_returns_parsed_options_full(self) -> None:
        parsed = parse_registration_options_json(
            {
                "rp": {"name": "Example Co", "id": "example.com"},
                "user": {"id": "AQIDBA", "name": "lee", "displayName": "Lee"},
                "challenge": "AQIDBAUGBwgJAA",
                "pubKeyCredParams": [
                    {"type": "public-key", "alg": -7},
                    {"type": "public-key", "alg": -8},
                    {"type": "public-key", "alg": -36},
                    {"type": "public-key", "alg": -37},
                    {"type": "public-key", "alg": -38},
                    {"type": "public-key", "alg": -39},
                    {"type": "public-key", "alg": -257},
                    {"type": "public-key", "alg": -258},
                    {"type": "public-key", "alg": -259},
                ],
                "timeout": 12000,
                "excludeCredentials": [
                    {
                        "id": "MTIzNDU2Nzg5MA",
                        "type": "public-key",
                        "transports": ["internal", "hybrid"],
                    }
                ],
                "authenticatorSelection": {
                    "authenticatorAttachment": "platform",
                    "residentKey": "required",
                    "requireResidentKey": True,
                    "userVerification": "discouraged",
                },
                "attestation": "direct",
                "hints": ["security-key", "client-device", "hybrid"],
            }
        )

        self.assertEqual(
            parsed.rp, PublicKeyCredentialRpEntity(id="example.com", name="Example Co")
        )
        self.assertEqual(
            parsed.user,
            PublicKeyCredentialUserEntity(
                id=base64url_to_bytes("AQIDBA"),
                name="lee",
                display_name="Lee",
            ),
        )
        self.assertEqual(parsed.attestation, AttestationConveyancePreference.DIRECT)
        self.assertEqual(
            parsed.authenticator_selection,
            AuthenticatorSelectionCriteria(
                authenticator_attachment=AuthenticatorAttachment.PLATFORM,
                resident_key=ResidentKeyRequirement.REQUIRED,
                require_resident_key=True,
                user_verification=UserVerificationRequirement.DISCOURAGED,
            ),
        )
        self.assertEqual(parsed.challenge, base64url_to_bytes("AQIDBAUGBwgJAA"))
        self.assertEqual(
            parsed.exclude_credentials,
            [
                PublicKeyCredentialDescriptor(
                    id=base64url_to_bytes("MTIzNDU2Nzg5MA"),
                    transports=[AuthenticatorTransport.INTERNAL, AuthenticatorTransport.HYBRID],
                )
            ],
        )
        self.assertEqual(
            parsed.pub_key_cred_params,
            [
                PublicKeyCredentialParameters(
                    alg=COSEAlgorithmIdentifier.ECDSA_SHA_256,
                    type="public-key",
                ),
                PublicKeyCredentialParameters(
                    alg=COSEAlgorithmIdentifier.EDDSA,
                    type="public-key",
                ),
                PublicKeyCredentialParameters(
                    alg=COSEAlgorithmIdentifier.ECDSA_SHA_512,
                    type="public-key",
                ),
                PublicKeyCredentialParameters(
                    alg=COSEAlgorithmIdentifier.RSASSA_PSS_SHA_256,
                    type="public-key",
                ),
                PublicKeyCredentialParameters(
                    alg=COSEAlgorithmIdentifier.RSASSA_PSS_SHA_384,
                    type="public-key",
                ),
                PublicKeyCredentialParameters(
                    alg=COSEAlgorithmIdentifier.RSASSA_PSS_SHA_512,
                    type="public-key",
                ),
                PublicKeyCredentialParameters(
                    alg=COSEAlgorithmIdentifier.RSASSA_PKCS1_v1_5_SHA_256,
                    type="public-key",
                ),
                PublicKeyCredentialParameters(
                    alg=COSEAlgorithmIdentifier.RSASSA_PKCS1_v1_5_SHA_384,
                    type="public-key",
                ),
                PublicKeyCredentialParameters(
                    alg=COSEAlgorithmIdentifier.RSASSA_PKCS1_v1_5_SHA_512,
                    type="public-key",
                ),
            ],
        )
        self.assertEqual(parsed.timeout, 12000)
        self.assertEqual(
            parsed.hints,
            [
                PublicKeyCredentialHint.SECURITY_KEY,
                PublicKeyCredentialHint.CLIENT_DEVICE,
                PublicKeyCredentialHint.HYBRID,
            ],
        )

    def test_supports_json_string(self) -> None:
        parsed = parse_registration_options_json(
            '{"rp": {"name": "Example Co", "id": "example.com"}, "user": {"id": "vEC5nFXSxpc_W68bX59JeD3c_-1XDJ5RblcWjY3Tx7RvfC0rkB19UWadf6wDEWG8T1ztksOYMim0sJIn6z_5tw", "name": "bob", "displayName": "bob"}, "challenge": "scb_z5GweYijAT2ppsB0HAklsw96fPs_tOWh-myqkOeb9rcvhWBwUZ56J3t3eocgjHkS4Mf3XeXTOQc1ySvk5w", "authenticatorSelection": {"userVerification": "required"}, "pubKeyCredParams": [{"type": "public-key", "alg": -36}], "timeout": 60000, "excludeCredentials": [], "attestation": "none"}'
        )

        self.assertEqual(
            parsed.rp, PublicKeyCredentialRpEntity(id="example.com", name="Example Co")
        )
        self.assertEqual(
            parsed.user,
            PublicKeyCredentialUserEntity(
                id=base64url_to_bytes(
                    "vEC5nFXSxpc_W68bX59JeD3c_-1XDJ5RblcWjY3Tx7RvfC0rkB19UWadf6wDEWG8T1ztksOYMim0sJIn6z_5tw"
                ),
                name="bob",
                display_name="bob",
            ),
        )
        self.assertEqual(parsed.attestation, AttestationConveyancePreference.NONE)
        self.assertEqual(
            parsed.authenticator_selection,
            AuthenticatorSelectionCriteria(user_verification=UserVerificationRequirement.REQUIRED),
        )
        self.assertEqual(
            parsed.challenge,
            base64url_to_bytes(
                "scb_z5GweYijAT2ppsB0HAklsw96fPs_tOWh-myqkOeb9rcvhWBwUZ56J3t3eocgjHkS4Mf3XeXTOQc1ySvk5w"
            ),
        )
        self.assertEqual(parsed.exclude_credentials, [])
        self.assertEqual(
            parsed.pub_key_cred_params,
            [
                PublicKeyCredentialParameters(
                    alg=COSEAlgorithmIdentifier.ECDSA_SHA_512,
                    type="public-key",
                )
            ],
        )
        self.assertEqual(parsed.timeout, 60000)

    def test_supports_options_to_json_output(self) -> None:
        """
        Test that output from `generate_registration_options()` that's fed directly into
        `options_to_json()` gets parsed back into the original options without any changes along
        the way.
        """
        opts = generate_registration_options(
            rp_id="example.com",
            rp_name="Example Co",
            user_id=bytes([1, 2, 3, 4]),
            user_name="lee",
            user_display_name="Lee",
            attestation=AttestationConveyancePreference.DIRECT,
            authenticator_selection=AuthenticatorSelectionCriteria(
                authenticator_attachment=AuthenticatorAttachment.PLATFORM,
                resident_key=ResidentKeyRequirement.REQUIRED,
                require_resident_key=True,
                user_verification=UserVerificationRequirement.DISCOURAGED,
            ),
            challenge=bytes([1, 2, 3, 4, 5, 6, 7, 8, 9, 0]),
            exclude_credentials=[
                PublicKeyCredentialDescriptor(
                    id=b"1234567890",
                    transports=[AuthenticatorTransport.INTERNAL, AuthenticatorTransport.HYBRID],
                ),
            ],
            supported_pub_key_algs=[COSEAlgorithmIdentifier.ECDSA_SHA_512],
            timeout=12000,
            hints=[
                PublicKeyCredentialHint.CLIENT_DEVICE,
                PublicKeyCredentialHint.SECURITY_KEY,
                PublicKeyCredentialHint.HYBRID,
            ],
        )

        opts_json = options_to_json(opts)

        parsed_opts_json = parse_registration_options_json(opts_json)

        self.assertEqual(parsed_opts_json.rp, opts.rp)
        self.assertEqual(parsed_opts_json.user, opts.user)
        self.assertEqual(parsed_opts_json.attestation, opts.attestation)
        self.assertEqual(parsed_opts_json.authenticator_selection, opts.authenticator_selection)
        self.assertEqual(parsed_opts_json.challenge, opts.challenge)
        self.assertEqual(parsed_opts_json.exclude_credentials, opts.exclude_credentials)
        self.assertEqual(parsed_opts_json.pub_key_cred_params, opts.pub_key_cred_params)
        self.assertEqual(parsed_opts_json.timeout, opts.timeout)
        self.assertEqual(parsed_opts_json.hints, opts.hints)

    def test_raises_on_non_dict_json(self) -> None:
        with self.assertRaisesRegex(InvalidJSONStructure, "not a JSON object"):
            parse_registration_options_json("[0]")

    def test_raises_on_missing_rp(self) -> None:
        with self.assertRaisesRegex(InvalidJSONStructure, "missing required rp"):
            parse_registration_options_json({})

    def test_raises_on_malformed_rp_id(self) -> None:
        with self.assertRaisesRegex(InvalidJSONStructure, "id present but not string"):
            parse_registration_options_json(
                {
                    "rp": {"id": 0},
                }
            )

    def test_raises_on_missing_missing_rp_name(self) -> None:
        with self.assertRaisesRegex(InvalidJSONStructure, "missing required name"):
            parse_registration_options_json(
                {
                    "rp": {"id": "example.com"},
                }
            )

    def test_raises_on_missing_user(self) -> None:
        with self.assertRaisesRegex(InvalidJSONStructure, "missing required user"):
            parse_registration_options_json(
                {
                    "rp": {"id": "example.com", "name": "Example Co"},
                }
            )

    def test_raises_on_missing_user_id(self) -> None:
        with self.assertRaisesRegex(InvalidJSONStructure, "missing required id"):
            parse_registration_options_json(
                {
                    "rp": {"id": "example.com", "name": "Example Co"},
                    "user": {},
                }
            )

    def test_raises_on_missing_user_name(self) -> None:
        with self.assertRaisesRegex(InvalidJSONStructure, "missing required name"):
            parse_registration_options_json(
                {
                    "rp": {"id": "example.com", "name": "Example Co"},
                    "user": {"id": "aaaa"},
                }
            )

    def test_raises_on_missing_user_display_name(self) -> None:
        with self.assertRaisesRegex(InvalidJSONStructure, "missing required displayName"):
            parse_registration_options_json(
                {
                    "rp": {"id": "example.com", "name": "Example Co"},
                    "user": {"id": "aaaa", "name": "Lee"},
                }
            )

    def test_raises_on_missing_attestation(self) -> None:
        with self.assertRaisesRegex(InvalidJSONStructure, "missing required attestation"):
            parse_registration_options_json(
                {
                    "rp": {"id": "example.com", "name": "Example Co"},
                    "user": {"id": "aaaa", "name": "lee", "displayName": "Lee"},
                }
            )

    def test_raises_on_unrecognized_attestation(self) -> None:
        with self.assertRaisesRegex(InvalidJSONStructure, "attestation was invalid value"):
            parse_registration_options_json(
                {
                    "rp": {"id": "example.com", "name": "Example Co"},
                    "user": {"id": "aaaa", "name": "lee", "displayName": "Lee"},
                    "attestation": "if_you_feel_like_it",
                }
            )

    def test_supports_optional_authenticator_selection(self) -> None:
        opts = parse_registration_options_json(
            {
                "rp": {"id": "example.com", "name": "Example Co"},
                "user": {"id": "aaaa", "name": "lee", "displayName": "Lee"},
                "attestation": "none",
                "challenge": "AAAA",
                "pubKeyCredParams": [{"type": "public-key", "alg": -7}],
            }
        )

        self.assertIsNone(opts.authenticator_selection)

    def test_raises_on_invalid_authenticator_selection_authenticator_attachment(self) -> None:
        with self.assertRaisesRegex(InvalidJSONStructure, "attachment was invalid value"):
            parse_registration_options_json(
                {
                    "rp": {"id": "example.com", "name": "Example Co"},
                    "user": {"id": "aaaa", "name": "lee", "displayName": "Lee"},
                    "attestation": "none",
                    "authenticatorSelection": {"authenticatorAttachment": "pcie"},
                }
            )

    def test_raises_on_invalid_authenticator_selection_resident_key(self) -> None:
        with self.assertRaisesRegex(InvalidJSONStructure, "residentKey was invalid value"):
            parse_registration_options_json(
                {
                    "rp": {"id": "example.com", "name": "Example Co"},
                    "user": {"id": "aaaa", "name": "lee", "displayName": "Lee"},
                    "attestation": "none",
                    "authenticatorSelection": {"residentKey": "yes_please"},
                }
            )

    def test_raises_on_invalid_authenticator_selection_require_resident_key(self) -> None:
        with self.assertRaisesRegex(
            InvalidJSONStructure, "requireResidentKey was invalid boolean"
        ):
            parse_registration_options_json(
                {
                    "rp": {"id": "example.com", "name": "Example Co"},
                    "user": {"id": "aaaa", "name": "lee", "displayName": "Lee"},
                    "attestation": "none",
                    "authenticatorSelection": {"requireResidentKey": "always"},
                }
            )

    def test_raises_on_invalid_authenticator_selection_user_verification(self) -> None:
        with self.assertRaisesRegex(InvalidJSONStructure, "userVerification was invalid value"):
            parse_registration_options_json(
                {
                    "rp": {"id": "example.com", "name": "Example Co"},
                    "user": {"id": "aaaa", "name": "lee", "displayName": "Lee"},
                    "attestation": "none",
                    "authenticatorSelection": {"userVerification": "when_inconvenient"},
                }
            )

    def test_raises_on_missing_challenge(self) -> None:
        with self.assertRaisesRegex(InvalidJSONStructure, "missing required challenge"):
            parse_registration_options_json(
                {
                    "rp": {"id": "example.com", "name": "Example Co"},
                    "user": {"id": "aaaa", "name": "lee", "displayName": "Lee"},
                    "attestation": "none",
                }
            )

    def test_raises_on_missing_pub_key_cred_params(self) -> None:
        with self.assertRaisesRegex(InvalidJSONStructure, "pubKeyCredParams was invalid value"):
            parse_registration_options_json(
                {
                    "rp": {"id": "example.com", "name": "Example Co"},
                    "user": {"id": "aaaa", "name": "lee", "displayName": "Lee"},
                    "attestation": "none",
                    "challenge": "aaaa",
                }
            )

    def test_raises_on_pub_key_cred_params_entry_with_invalid_alg(self) -> None:
        with self.assertRaisesRegex(InvalidJSONStructure, "entry had invalid alg"):
            parse_registration_options_json(
                {
                    "rp": {"id": "example.com", "name": "Example Co"},
                    "user": {"id": "aaaa", "name": "lee", "displayName": "Lee"},
                    "attestation": "none",
                    "challenge": "aaaa",
                    "pubKeyCredParams": [{"alg": 0}],
                }
            )

    def test_supports_optional_exclude_credentials(self) -> None:
        opts = parse_registration_options_json(
            {
                "rp": {"id": "example.com", "name": "Example Co"},
                "user": {"id": "aaaa", "name": "lee", "displayName": "Lee"},
                "attestation": "none",
                "challenge": "aaaa",
                "pubKeyCredParams": [{"alg": -7}],
            }
        )

        self.assertIsNone(opts.exclude_credentials)

    def test_raises_on_exclude_credentials_entry_missing_id(self) -> None:
        with self.assertRaisesRegex(InvalidJSONStructure, "missing required id"):
            parse_registration_options_json(
                {
                    "rp": {"id": "example.com", "name": "Example Co"},
                    "user": {"id": "aaaa", "name": "lee", "displayName": "Lee"},
                    "attestation": "none",
                    "challenge": "aaaa",
                    "pubKeyCredParams": [{"alg": -7}],
                    "excludeCredentials": [{}],
                }
            )

    def test_raises_on_exclude_credentials_entry_invalid_transports(self) -> None:
        with self.assertRaisesRegex(InvalidJSONStructure, "transports was not list"):
            parse_registration_options_json(
                {
                    "rp": {"id": "example.com", "name": "Example Co"},
                    "user": {"id": "aaaa", "name": "lee", "displayName": "Lee"},
                    "attestation": "none",
                    "challenge": "aaaa",
                    "pubKeyCredParams": [{"alg": -7}],
                    "excludeCredentials": [{"id": "aaaa", "transports": ""}],
                }
            )

    def test_raises_on_exclude_credentials_entry_invalid_transports_entry(self) -> None:
        with self.assertRaisesRegex(InvalidJSONStructure, "entry transports had invalid value"):
            parse_registration_options_json(
                {
                    "rp": {"id": "example.com", "name": "Example Co"},
                    "user": {"id": "aaaa", "name": "lee", "displayName": "Lee"},
                    "attestation": "none",
                    "challenge": "aaaa",
                    "pubKeyCredParams": [{"alg": -7}],
                    "excludeCredentials": [{"id": "aaaa", "transports": ["pcie"]}],
                }
            )

    def test_supports_missing_timeout(self) -> None:
        opts = parse_registration_options_json(
            {
                "rp": {"id": "example.com", "name": "Example Co"},
                "user": {"id": "aaaa", "name": "lee", "displayName": "Lee"},
                "attestation": "none",
                "challenge": "aaaa",
                "pubKeyCredParams": [{"alg": -7}],
            }
        )

        self.assertIsNone(opts.timeout)

    def test_supports_empty_hints(self) -> None:
        opts = parse_registration_options_json(
            {
                "rp": {"id": "example.com", "name": "Example Co"},
                "user": {"id": "aaaa", "name": "lee", "displayName": "Lee"},
                "attestation": "none",
                "challenge": "aaaa",
                "pubKeyCredParams": [{"alg": -7}],
                "hints": [],
            }
        )

        self.assertEqual(opts.hints, [])

    def test_raises_on_invalid_hints_assignment(self) -> None:
        with self.assertRaisesRegex(InvalidJSONStructure, "hints was invalid value"):
            parse_registration_options_json(
                {
                    "rp": {"id": "example.com", "name": "Example Co"},
                    "user": {"id": "aaaa", "name": "lee", "displayName": "Lee"},
                    "attestation": "none",
                    "challenge": "aaaa",
                    "pubKeyCredParams": [{"alg": -7}],
                    "hints": "security-key",
                }
            )

    def test_raises_on_invalid_hints_entry(self) -> None:
        with self.assertRaisesRegex(InvalidJSONStructure, "hints had invalid value"):
            parse_registration_options_json(
                {
                    "rp": {"id": "example.com", "name": "Example Co"},
                    "user": {"id": "aaaa", "name": "lee", "displayName": "Lee"},
                    "attestation": "none",
                    "challenge": "aaaa",
                    "pubKeyCredParams": [{"alg": -7}],
                    "hints": ["platform"],
                }
            )

    def test_supports_optional_hints(self) -> None:
        opts = parse_registration_options_json(
            {
                "rp": {"id": "example.com", "name": "Example Co"},
                "user": {"id": "aaaa", "name": "lee", "displayName": "Lee"},
                "attestation": "none",
                "challenge": "aaaa",
                "pubKeyCredParams": [{"alg": -7}],
            }
        )

        self.assertIsNone(opts.hints)



from typing import List, Optional

from webauthn.helpers import generate_challenge
from webauthn.helpers.structs import (
    PublicKeyCredentialDescriptor,
    PublicKeyCredentialRequestOptions,
    UserVerificationRequirement,
)


def generate_authentication_options(
    *,
    rp_id: str,
    challenge: Optional[bytes] = None,
    timeout: int = 60000,
    allow_credentials: Optional[List[PublicKeyCredentialDescriptor]] = None,
    user_verification: UserVerificationRequirement = UserVerificationRequirement.PREFERRED,
) -> PublicKeyCredentialRequestOptions:
    """Generate options for retrieving a credential via navigator.credentials.get()

    Args:
        `rp_id`: The Relying Party's unique identifier as specified in attestations.
        (optional) `challenge`: A byte sequence for the authenticator to return back in its response. Defaults to 64 random bytes.
        (optional) `timeout`: How long in milliseconds the browser should give the user to choose an authenticator. This value is a *hint* and may be ignored by the browser.
        (optional) `allow_credentials`: A list of credentials registered to the user.
        (optional) `user_verification`: The RP's preference for the authenticator's enforcement of the "user verified" flag.

    Returns:
        Authentication options ready for the browser. Consider using `helpers.options_to_json()` in this library to quickly convert the options to JSON.
    """

    if not rp_id:
        raise ValueError("rp_id cannot be an empty string")

    ########
    # Set defaults for required values
    ########

    if not challenge:
        challenge = generate_challenge()

    if not allow_credentials:
        allow_credentials = []

    return PublicKeyCredentialRequestOptions(
        rp_id=rp_id,
        challenge=challenge,
        timeout=timeout,
        allow_credentials=allow_credentials,
        user_verification=user_verification,
    )



import secrets


def generate_challenge() -> bytes:
    """
    Create a random value for the authenticator to sign, going above and beyond the recommended
    number of random bytes as per https://www.w3.org/TR/webauthn-2/#sctn-cryptographic-challenges:

    "In order to prevent replay attacks, the challenges MUST contain enough entropy to make
    guessing them infeasible. Challenges SHOULD therefore be at least 16 bytes long."
    """
    return secrets.token_bytes(64)



import hashlib
from typing import Optional

from .cose import COSEAlgorithmIdentifier

SHA_256 = [
    COSEAlgorithmIdentifier.ECDSA_SHA_256,
    COSEAlgorithmIdentifier.RSASSA_PSS_SHA_256,
    COSEAlgorithmIdentifier.RSASSA_PKCS1_v1_5_SHA_256,
]
SHA_384 = [
    COSEAlgorithmIdentifier.RSASSA_PSS_SHA_384,
    COSEAlgorithmIdentifier.RSASSA_PKCS1_v1_5_SHA_384,
]
SHA_512 = [
    COSEAlgorithmIdentifier.ECDSA_SHA_512,
    COSEAlgorithmIdentifier.RSASSA_PSS_SHA_512,
    COSEAlgorithmIdentifier.RSASSA_PKCS1_v1_5_SHA_512,
]
SHA_1 = [
    COSEAlgorithmIdentifier.RSASSA_PKCS1_v1_5_SHA_1,
]


def hash_by_alg(to_hash: bytes, alg: Optional[COSEAlgorithmIdentifier] = None) -> bytes:
    """
    Generate a hash of `to_hash` by the specified COSE algorithm ID. Defaults to hashing
    with SHA256
    """
    # Default to SHA256 for hashing
    hash = hashlib.sha256()

    if alg in SHA_384:
        hash = hashlib.sha384()
    elif alg in SHA_512:
        hash = hashlib.sha512()
    elif alg in SHA_1:
        hash = hashlib.sha1()

    hash.update(to_hash)
    return hash.digest()



from unittest import TestCase

from webauthn.helpers.snake_case_to_camel_case import snake_case_to_camel_case


class TestWebAuthnSnakeCaseToCamelCase(TestCase):
    def test_converts_snake_case_to_camel_case(self) -> None:
        output = snake_case_to_camel_case("snake_case")

        assert output == "snakeCase"

    def test_converts_client_data_json(self) -> None:
        output = snake_case_to_camel_case("client_data_json")

        assert output == "clientDataJSON"



from .aaguid_to_string import aaguid_to_string
from .base64url_to_bytes import base64url_to_bytes
from .bytes_to_base64url import bytes_to_base64url
from .byteslike_to_bytes import byteslike_to_bytes
from .decode_credential_public_key import decode_credential_public_key
from .decoded_public_key_to_cryptography import decoded_public_key_to_cryptography
from .encode_cbor import encode_cbor
from .generate_challenge import generate_challenge
from .generate_user_handle import generate_user_handle
from .hash_by_alg import hash_by_alg
from .options_to_json import options_to_json
from .parse_attestation_object import parse_attestation_object
from .parse_authentication_credential_json import parse_authentication_credential_json
from .parse_authentication_options_json import parse_authentication_options_json
from .parse_authenticator_data import parse_authenticator_data
from .parse_backup_flags import parse_backup_flags
from .parse_cbor import parse_cbor
from .parse_client_data_json import parse_client_data_json
from .parse_registration_credential_json import parse_registration_credential_json
from .parse_registration_options_json import parse_registration_options_json
from .validate_certificate_chain import validate_certificate_chain
from .verify_safetynet_timestamp import verify_safetynet_timestamp
from .verify_signature import verify_signature

__all__ = [
    "aaguid_to_string",
    "base64url_to_bytes",
    "bytes_to_base64url",
    "byteslike_to_bytes",
    "decode_credential_public_key",
    "decoded_public_key_to_cryptography",
    "encode_cbor",
    "generate_challenge",
    "generate_user_handle",
    "hash_by_alg",
    "options_to_json",
    "parse_attestation_object",
    "parse_authenticator_data",
    "parse_authentication_credential_json",
    "parse_authentication_options_json",
    "parse_backup_flags",
    "parse_cbor",
    "parse_client_data_json",
    "parse_registration_credential_json",
    "parse_registration_options_json",
    "validate_certificate_chain",
    "verify_safetynet_timestamp",
    "verify_signature",
]



from unittest import TestCase
from unittest.mock import MagicMock, patch

from webauthn import generate_authentication_options
from webauthn.helpers.structs import (
    PublicKeyCredentialDescriptor,
    UserVerificationRequirement,
)


class TestWebAuthnGenerateAttestationOptions(TestCase):
    @patch("secrets.token_bytes")
    def test_generates_options_with_defaults(self, token_bytes_mock: MagicMock) -> None:
        token_bytes_mock.return_value = b"12345"

        options = generate_authentication_options(rp_id="example.com")

        assert options.challenge == b"12345"
        assert options.timeout == 60000
        assert options.rp_id == "example.com"
        assert options.allow_credentials == []
        assert options.user_verification == UserVerificationRequirement.PREFERRED

    def test_generates_options_with_custom_values(self) -> None:
        options = generate_authentication_options(
            rp_id="example.com",
            allow_credentials=[
                PublicKeyCredentialDescriptor(id=b"12345"),
            ],
            challenge=b"this_is_a_challenge",
            timeout=12000,
            user_verification=UserVerificationRequirement.REQUIRED,
        )

        assert options.challenge == b"this_is_a_challenge"
        assert options.timeout == 12000
        assert options.rp_id == "example.com"
        assert options.allow_credentials == [
            PublicKeyCredentialDescriptor(id=b"12345"),
        ]
        assert options.user_verification == UserVerificationRequirement.REQUIRED

    def test_raises_on_empty_rp_id(self) -> None:
        with self.assertRaisesRegex(ValueError, "rp_id"):
            generate_authentication_options(
                rp_id="",
            )



from unittest import TestCase

from webauthn.helpers import base64url_to_bytes
from webauthn.helpers.cose import COSEAlgorithmIdentifier
from webauthn.helpers.parse_attestation_object import parse_attestation_object
from webauthn.helpers.structs import AttestationFormat


class TestParseAttestationObject(TestCase):
    """
    To generate values to pass into `parse_attestation_object()`:

    1. Perform an attestation in the browser until you receive a value back from
       `navigator.credentials.create()`
    2. Base64URL-encode the value of `response.attestationObject` to make it easy to
       include in tests below
    """

    def test_can_parse_good_none_attestation_object(self) -> None:
        attestation_object = base64url_to_bytes(
            "o2NmbXRkbm9uZWdhdHRTdG10oGhhdXRoRGF0YVjESZYN5YgOjGh0NBcPZHZgW4_krrmihjLHmVzzuoMdl2NFAAAAKwAAAAAAAAAAAAAAAAAAAAAAQBGidLzwQ3gFpfrQexcVIM1EKISJJUorv8HP1BQ-Km2G7Lw5VyGWRVG7pibkOG3OJ2LYG9tZY57jUrSjOqfEJgilAQIDJiABIVggBlNr3gyj0tq3QJrYd7t74BncruqqAVPCDlMfaU-7_kIiWCCXa5Oxqay8XoVwsTYzLSDp-ULUIAalnTyiQJrrrRWB_A"
        )

        output = parse_attestation_object(attestation_object)

        assert output.fmt == AttestationFormat.NONE
        # attestation statement
        att_stmt = output.att_stmt
        assert att_stmt.sig is None
        assert att_stmt.x5c is None
        assert att_stmt.response is None
        assert att_stmt.alg is None
        assert att_stmt.ver is None
        assert att_stmt.cert_info is None
        assert att_stmt.pub_area is None
        # authenticator data
        auth_data = output.auth_data
        assert auth_data.rp_id_hash == base64url_to_bytes(
            "SZYN5YgOjGh0NBcPZHZgW4_krrmihjLHmVzzuoMdl2M"
        )
        assert auth_data.flags.up
        assert auth_data.flags.uv
        assert auth_data.flags.at
        assert not auth_data.flags.ed
        assert auth_data.sign_count == 43
        # attested credential data
        assert auth_data.attested_credential_data
        attested_cred_data = auth_data.attested_credential_data
        assert attested_cred_data.aaguid == base64url_to_bytes("AAAAAAAAAAAAAAAAAAAAAA")
        assert attested_cred_data.credential_id == base64url_to_bytes(
            "EaJ0vPBDeAWl-tB7FxUgzUQohIklSiu_wc_UFD4qbYbsvDlXIZZFUbumJuQ4bc4nYtgb21ljnuNStKM6p8QmCA"
        )
        assert attested_cred_data.credential_public_key == base64url_to_bytes(
            "pQECAyYgASFYIAZTa94Mo9Lat0Ca2He7e-AZ3K7qqgFTwg5TH2lPu_5CIlggl2uTsamsvF6FcLE2My0g6flC1CAGpZ08okCa660Vgfw"
        )
        # extensions
        assert not auth_data.extensions

    def test_can_parse_good_packed_attestation_object(self) -> None:
        attestation_object = base64url_to_bytes(
            "o2NmbXRmcGFja2VkZ2F0dFN0bXSjY2FsZyZjc2lnWEcwRQIgRpuZ6hdaLAgWgCFTIo4BGSTBAxwwqk4u3s1-JAzv_H4CIQCZnfoic34aOwlac1A09eflEtb0V1kO7yGhHOw5P5wVWmN4NWOBWQLBMIICvTCCAaWgAwIBAgIEKudiYzANBgkqhkiG9w0BAQsFADAuMSwwKgYDVQQDEyNZdWJpY28gVTJGIFJvb3QgQ0EgU2VyaWFsIDQ1NzIwMDYzMTAgFw0xNDA4MDEwMDAwMDBaGA8yMDUwMDkwNDAwMDAwMFowbjELMAkGA1UEBhMCU0UxEjAQBgNVBAoMCVl1YmljbyBBQjEiMCAGA1UECwwZQXV0aGVudGljYXRvciBBdHRlc3RhdGlvbjEnMCUGA1UEAwweWXViaWNvIFUyRiBFRSBTZXJpYWwgNzE5ODA3MDc1MFkwEwYHKoZIzj0CAQYIKoZIzj0DAQcDQgAEKgOGXmBD2Z4R_xCqJVRXhL8Jr45rHjsyFykhb1USGozZENOZ3cdovf5Ke8fj2rxi5tJGn_VnW4_6iQzKdIaeP6NsMGowIgYJKwYBBAGCxAoCBBUxLjMuNi4xLjQuMS40MTQ4Mi4xLjEwEwYLKwYBBAGC5RwCAQEEBAMCBDAwIQYLKwYBBAGC5RwBAQQEEgQQbUS6m_bsLkm5MAyP6SDLczAMBgNVHRMBAf8EAjAAMA0GCSqGSIb3DQEBCwUAA4IBAQByV9A83MPhFWmEkNb4DvlbUwcjc9nmRzJjKxHc3HeK7GvVkm0H4XucVDB4jeMvTke0WHb_jFUiApvpOHh5VyMx5ydwFoKKcRs5x0_WwSWL0eTZ5WbVcHkDR9pSNcA_D_5AsUKOBcbpF5nkdVRxaQHuuIuwV4k1iK2IqtMNcU8vL6w21U261xCcWwJ6sMq4zzVO8QCKCQhsoIaWrwz828GDmPzfAjFsJiLJXuYivdHACkeJ5KHMt0mjVLpfJ2BCML7_rgbmvwL7wBW80VHfNdcKmKjkLcpEiPzwcQQhiN_qHV90t-p4iyr5xRSpurlP5zic2hlRkLKxMH2_kRjhqSn4aGF1dGhEYXRhWMRJlg3liA6MaHQ0Fw9kdmBbj-SuuaKGMseZXPO6gx2XY0UAAAAqbUS6m_bsLkm5MAyP6SDLcwBAFsWBrFcw8yRjxV8z18Egh91o1AScNRYkIuUoY6wIlIhslDpP7eydKi1q5s9g1ugDP9mqBlPDDFPRbH6YLwHbtqUBAgMmIAEhWCAq3y0RWh8nLzanBZQwTA7yAbUy9KEDAM0b3N9Elrb0VCJYIJrX7ygtpyInb5mXBE7g9YEow6xWrJ400HhL2r4q5tzV",
        )

        output = parse_attestation_object(attestation_object)

        assert output.fmt == AttestationFormat.PACKED
        # attestation statement
        att_stmt = output.att_stmt
        assert att_stmt.sig and att_stmt.sig == base64url_to_bytes(
            "MEUCIEabmeoXWiwIFoAhUyKOARkkwQMcMKpOLt7NfiQM7_x-AiEAmZ36InN-GjsJWnNQNPXn5RLW9FdZDu8hoRzsOT-cFVo"
        )
        assert att_stmt.x5c and len(att_stmt.x5c) == 1
        assert att_stmt.response is None
        assert att_stmt.alg == COSEAlgorithmIdentifier.ECDSA_SHA_256
        assert att_stmt.ver is None
        assert att_stmt.cert_info is None
        assert att_stmt.pub_area is None
        # authenticator data
        auth_data = output.auth_data
        assert auth_data.rp_id_hash == base64url_to_bytes(
            "SZYN5YgOjGh0NBcPZHZgW4_krrmihjLHmVzzuoMdl2M"
        )
        assert auth_data.flags.up
        assert auth_data.flags.uv
        assert auth_data.flags.at
        assert not auth_data.flags.ed
        assert auth_data.sign_count == 42
        # attested credential data
        assert auth_data.attested_credential_data
        attested_cred_data = auth_data.attested_credential_data
        assert attested_cred_data.aaguid == base64url_to_bytes("bUS6m_bsLkm5MAyP6SDLcw")
        assert attested_cred_data.credential_id == base64url_to_bytes(
            "FsWBrFcw8yRjxV8z18Egh91o1AScNRYkIuUoY6wIlIhslDpP7eydKi1q5s9g1ugDP9mqBlPDDFPRbH6YLwHbtg"
        )
        assert attested_cred_data.credential_public_key == base64url_to_bytes(
            "pQECAyYgASFYICrfLRFaHycvNqcFlDBMDvIBtTL0oQMAzRvc30SWtvRUIlggmtfvKC2nIidvmZcETuD1gSjDrFasnjTQeEvavirm3NU"
        )
        # extensions
        assert auth_data.extensions is None



from unittest import TestCase
from unittest.mock import patch
from datetime import datetime
from OpenSSL.crypto import X509, X509Store, X509StoreContextError

from webauthn.helpers.exceptions import InvalidCertificateChain
from webauthn.helpers.known_root_certs import (
    apple_webauthn_root_ca,
    globalsign_root_ca,
)
from webauthn.helpers.validate_certificate_chain import (
    validate_certificate_chain,
)

from .helpers.x509store import patch_validate_certificate_chain_x509store_getter

apple_x5c_certs = [
    # 2021-08-31 @ 23:02:07Z <-> 2021-09-03 @ 23:02:07Z
    bytes.fromhex(
        "30820243308201c9a0030201020206017ba3992221300a06082a8648ce3d0403023048311c301a06035504030c134170706c6520576562417574686e204341203131133011060355040a0c0a4170706c6520496e632e3113301106035504080c0a43616c69666f726e6961301e170d3231303833313233303230375a170d3231303930333233303230375a3081913149304706035504030c4062313066373138626335646437353838383661316438636662356238623633313732396634643765346261303639616230613939326331633038343738616639311a3018060355040b0c114141412043657274696669636174696f6e31133011060355040a0c0a4170706c6520496e632e3113301106035504080c0a43616c69666f726e69613059301306072a8648ce3d020106082a8648ce3d03010703420004d124b0e9ff8192723c9ee2fa4f8170d373e03286cf880aeec7008a14cdea64724963e05bb8c44a9f980ded12aa8a33795cf81d31e74116ced6f1f4c5eb0c358fa3553053300c0603551d130101ff04023000300e0603551d0f0101ff0404030204f0303306092a864886f76364080204263024a1220420e457e5bc292f1635210248ed2e776ba129c7cc469524a75356836caef2f058a0300a06082a8648ce3d0403020368003065023065c6e7075ddacb50879a8412904759013d0da78726408759a01f1994c1795a69c2c1d11306c2d1bc97be6141627b8677023100ab0b9e7d97ca2b603b1edb6e264c49bf1971380c2afa5d37f8c4ff5a5de6d457a19cb80c02b2edf94b0853e0482f8686"
    ),
    # 2020-03-18 @ 18:38:01Z <-> 2030-03-13 @ 00:00:00Z
    bytes.fromhex(
        "30820234308201baa003020102021056255395c7a7fb40ebe228d8260853b6300a06082a8648ce3d040303304b311f301d06035504030c164170706c6520576562417574686e20526f6f7420434131133011060355040a0c0a4170706c6520496e632e3113301106035504080c0a43616c69666f726e6961301e170d3230303331383138333830315a170d3330303331333030303030305a3048311c301a06035504030c134170706c6520576562417574686e204341203131133011060355040a0c0a4170706c6520496e632e3113301106035504080c0a43616c69666f726e69613076301006072a8648ce3d020106052b8104002203620004832e872f261491810225b9f5fcd6bb6378b5f55f3fcb045bc735993475fd549044df9bfe19211765c69a1dda050b38d45083401a434fb24d112d56c3e1cfbfcb9891fec0696081bef96cbc77c88dddaf46a5aee1dd515b5afaab93be9c0b2691a366306430120603551d130101ff040830060101ff020100301f0603551d2304183016801426d764d9c578c25a67d1a7de6b12d01b63f1c6d7301d0603551d0e04160414ebae82c4ffa1ac5b51d4cf24610500be63bd7788300e0603551d0f0101ff040403020106300a06082a8648ce3d0403030368003065023100dd8b1a3481a5fad9dbb4e7657b841e144c27b75b876a4186c2b1475750337227efe554457ef648950c632e5c483e70c102302c8a6044dc201fcfe59bc34d2930c1487851d960ed6a75f1eb4acabe38cd25b897d0c805bef0c7f78b07a571c6e80e07"
    ),
]


class TestValidateCertificateChain(TestCase):
    def setUp(self):
        # Setting the time to something that satisfies all these:
        # (Leaf) 20210831230207Z <-> 20210903230207Z <- Earliest expiration
        # (Int.) 20200318183801Z <-> 20300313000000Z
        # (Root) 20200318182132Z <-> 20450315000000Z
        self.x509store_time = datetime(2021, 9, 1, 0, 0, 0)

    @patch_validate_certificate_chain_x509store_getter
    def test_validates_certificate_chain(self, patched_x509store: X509Store) -> None:
        patched_x509store.set_time(self.x509store_time)

        try:
            validate_certificate_chain(
                x5c=apple_x5c_certs,
                pem_root_certs_bytes=[apple_webauthn_root_ca],
            )
        except Exception as err:
            print(err)
            self.fail("validate_certificate_chain failed when it should have succeeded")

    @patch_validate_certificate_chain_x509store_getter
    def test_throws_on_bad_root_cert(self, patched_x509store: X509Store) -> None:
        patched_x509store.set_time(self.x509store_time)

        with self.assertRaises(InvalidCertificateChain):
            validate_certificate_chain(
                x5c=apple_x5c_certs,
                # An obviously invalid root cert for these x5c certs
                pem_root_certs_bytes=[globalsign_root_ca],
            )

    def test_passes_on_no_root_certs(self):
        try:
            validate_certificate_chain(
                x5c=apple_x5c_certs,
            )
        except Exception as err:
            print(err)
            self.fail("validate_certificate_chain failed when it should have succeeded")

    def test_passes_on_empty_root_certs_array(self):
        try:
            validate_certificate_chain(
                x5c=apple_x5c_certs,
                pem_root_certs_bytes=[],
            )
        except Exception as err:
            print(err)
            self.fail("validate_certificate_chain failed when it should have succeeded")

    def test_includes_original_exception_when_raising(self):
        """
        A low-effort attempt at ensuring that the attempt to validate a certificate chain will
        include the original exception raised by the X509 validation library, instead of just its
        message.
        """
        custom_exception = X509StoreContextError(message="Oops", certificate=X509(), errors=[])

        with patch(
            "OpenSSL.crypto.X509StoreContext.verify_certificate", side_effect=custom_exception
        ):
            with self.assertRaises(InvalidCertificateChain) as ctx:
                validate_certificate_chain(
                    x5c=apple_x5c_certs,
                    pem_root_certs_bytes=[apple_webauthn_root_ca],
                )

            cause = ctx.exception.__cause__
            self.assertEqual(cause, custom_exception)



from unittest import TestCase

from webauthn.helpers.generate_challenge import generate_challenge


class TestWebAuthnGenerateChallenge(TestCase):
    def test_generates_byte_sequence(self) -> None:
        output = generate_challenge()

        self.assertEqual(type(output), bytes)
        self.assertEqual(len(output), 64)

    def test_generates_unique_value_each_time(self) -> None:
        output1 = generate_challenge()
        output2 = generate_challenge()

        self.assertNotEqual(output1, output2)



import json
from json.decoder import JSONDecodeError
from typing import Union

from .exceptions import InvalidAuthenticationResponse, InvalidJSONStructure
from .base64url_to_bytes import base64url_to_bytes
from .structs import (
    AuthenticationCredential,
    AuthenticatorAssertionResponse,
    AuthenticatorAttachment,
    PublicKeyCredentialType,
)


def parse_authentication_credential_json(json_val: Union[str, dict]) -> AuthenticationCredential:
    """
    Parse a JSON form of an authentication credential, as either a stringified JSON object or a
    plain dict, into an instance of AuthenticationCredential
    """
    if isinstance(json_val, str):
        try:
            json_val = json.loads(json_val)
        except JSONDecodeError:
            raise InvalidJSONStructure("Unable to decode credential as JSON")

    if not isinstance(json_val, dict):
        raise InvalidJSONStructure("Credential was not a JSON object")

    cred_id = json_val.get("id")
    if not isinstance(cred_id, str):
        raise InvalidJSONStructure("Credential missing required id")

    cred_raw_id = json_val.get("rawId")
    if not isinstance(cred_raw_id, str):
        raise InvalidJSONStructure("Credential missing required rawId")

    cred_response = json_val.get("response")
    if not isinstance(cred_response, dict):
        raise InvalidJSONStructure("Credential missing required response")

    response_client_data_json = cred_response.get("clientDataJSON")
    if not isinstance(response_client_data_json, str):
        raise InvalidJSONStructure("Credential response missing required clientDataJSON")

    response_authenticator_data = cred_response.get("authenticatorData")
    if not isinstance(response_authenticator_data, str):
        raise InvalidJSONStructure("Credential response missing required authenticatorData")

    response_signature = cred_response.get("signature")
    if not isinstance(response_signature, str):
        raise InvalidJSONStructure("Credential response missing required signature")

    cred_type = json_val.get("type")
    try:
        # Simply try to get the single matching Enum. We'll set the literal value below assuming
        # the code can get past here (this is basically a mypy optimization)
        PublicKeyCredentialType(cred_type)
    except ValueError as cred_type_exc:
        raise InvalidJSONStructure("Credential had unexpected type") from cred_type_exc

    response_user_handle = cred_response.get("userHandle")
    if isinstance(response_user_handle, str):
        # The `userHandle` string will most likely be base64url-encoded for ease of JSON
        # transmission as per the L3 Draft spec:
        # https://w3c.github.io/webauthn/#dictdef-authenticatorassertionresponsejson
        response_user_handle = base64url_to_bytes(response_user_handle)
    elif response_user_handle is not None:
        # If it's not a string, and it's not None, then it's definitely not valid
        raise InvalidJSONStructure("Credential response had unexpected userHandle")

    cred_authenticator_attachment = json_val.get("authenticatorAttachment")
    if isinstance(cred_authenticator_attachment, str):
        try:
            cred_authenticator_attachment = AuthenticatorAttachment(cred_authenticator_attachment)
        except ValueError as cred_attachment_exc:
            raise InvalidJSONStructure(
                "Credential had unexpected authenticatorAttachment"
            ) from cred_attachment_exc
    else:
        cred_authenticator_attachment = None

    try:
        authentication_credential = AuthenticationCredential(
            id=cred_id,
            raw_id=base64url_to_bytes(cred_raw_id),
            response=AuthenticatorAssertionResponse(
                client_data_json=base64url_to_bytes(response_client_data_json),
                authenticator_data=base64url_to_bytes(response_authenticator_data),
                signature=base64url_to_bytes(response_signature),
                user_handle=response_user_handle,
            ),
            authenticator_attachment=cred_authenticator_attachment,
            type=PublicKeyCredentialType.PUBLIC_KEY,
        )
    except Exception as exc:
        raise InvalidAuthenticationResponse(
            "Could not parse authentication credential from JSON data"
        ) from exc

    return authentication_credential



from unittest import TestCase

from webauthn.helpers.base64url_to_bytes import base64url_to_bytes


class TestWebAuthnBase64URLToBytes(TestCase):
    def test_converts_base64url_string_to_buffer(self) -> None:
        output = base64url_to_bytes("AQIDBAU")

        assert output == bytes([1, 2, 3, 4, 5])



from unittest import TestCase

from webauthn.helpers import base64url_to_bytes
from webauthn.helpers.structs import AttestationFormat
from webauthn import verify_registration_response


class TestVerifyRegistrationResponsePacked(TestCase):
    def test_verify_attestation_from_yubikey_firefox(self) -> None:
        credential = """{
            "id": "syGQPDZRUYdb4m3rdWeyPaIMYlbmydGp1TP_33vE_lqJ3PHNyTd0iKsnKr5WjnCcBzcesZrDEfB_RBLFzU3k4w",
            "rawId": "syGQPDZRUYdb4m3rdWeyPaIMYlbmydGp1TP_33vE_lqJ3PHNyTd0iKsnKr5WjnCcBzcesZrDEfB_RBLFzU3k4w",
            "response": {
                "attestationObject": "o2NmbXRmcGFja2VkZ2F0dFN0bXSjY2FsZyZjc2lnWEcwRQIhAOfrFlQpbavT6dJeTDJSCDzYSYPjBDHli2-syT2c1IiKAiAx5gQ2z5cHjdQX-jEHTb7JcjfQoVSW8fXszF5ihSgeOGN4NWOBWQLBMIICvTCCAaWgAwIBAgIEKudiYzANBgkqhkiG9w0BAQsFADAuMSwwKgYDVQQDEyNZdWJpY28gVTJGIFJvb3QgQ0EgU2VyaWFsIDQ1NzIwMDYzMTAgFw0xNDA4MDEwMDAwMDBaGA8yMDUwMDkwNDAwMDAwMFowbjELMAkGA1UEBhMCU0UxEjAQBgNVBAoMCVl1YmljbyBBQjEiMCAGA1UECwwZQXV0aGVudGljYXRvciBBdHRlc3RhdGlvbjEnMCUGA1UEAwweWXViaWNvIFUyRiBFRSBTZXJpYWwgNzE5ODA3MDc1MFkwEwYHKoZIzj0CAQYIKoZIzj0DAQcDQgAEKgOGXmBD2Z4R_xCqJVRXhL8Jr45rHjsyFykhb1USGozZENOZ3cdovf5Ke8fj2rxi5tJGn_VnW4_6iQzKdIaeP6NsMGowIgYJKwYBBAGCxAoCBBUxLjMuNi4xLjQuMS40MTQ4Mi4xLjEwEwYLKwYBBAGC5RwCAQEEBAMCBDAwIQYLKwYBBAGC5RwBAQQEEgQQbUS6m_bsLkm5MAyP6SDLczAMBgNVHRMBAf8EAjAAMA0GCSqGSIb3DQEBCwUAA4IBAQByV9A83MPhFWmEkNb4DvlbUwcjc9nmRzJjKxHc3HeK7GvVkm0H4XucVDB4jeMvTke0WHb_jFUiApvpOHh5VyMx5ydwFoKKcRs5x0_WwSWL0eTZ5WbVcHkDR9pSNcA_D_5AsUKOBcbpF5nkdVRxaQHuuIuwV4k1iK2IqtMNcU8vL6w21U261xCcWwJ6sMq4zzVO8QCKCQhsoIaWrwz828GDmPzfAjFsJiLJXuYivdHACkeJ5KHMt0mjVLpfJ2BCML7_rgbmvwL7wBW80VHfNdcKmKjkLcpEiPzwcQQhiN_qHV90t-p4iyr5xRSpurlP5zic2hlRkLKxMH2_kRjhqSn4aGF1dGhEYXRhWMRJlg3liA6MaHQ0Fw9kdmBbj-SuuaKGMseZXPO6gx2XY0UAAAA0bUS6m_bsLkm5MAyP6SDLcwBAsyGQPDZRUYdb4m3rdWeyPaIMYlbmydGp1TP_33vE_lqJ3PHNyTd0iKsnKr5WjnCcBzcesZrDEfB_RBLFzU3k46UBAgMmIAEhWCBAX_i3O3DvBnkGq_uLNk_PeAX5WwO_MIxBp0mhX6Lw7yJYIOW-1-Fch829McWvRUYAHTWZTx5IycKSGECL1UzUaK_8",
                "clientDataJSON": "eyJ0eXBlIjoid2ViYXV0aG4uY3JlYXRlIiwiY2hhbGxlbmdlIjoiOExCQ2lPWTNxMWNCWkhGQVd0UzRBWlpDaHpHcGh5NjdsSzdJNzB6S2k0eUM3cGdyUTJQY2g3bkFqTGsxd3E5Z3Jlc2hJQXNXMkFqaWJoWGpqSTBUbVEiLCJvcmlnaW4iOiJodHRwOi8vbG9jYWxob3N0OjUwMDAiLCJjcm9zc09yaWdpbiI6ZmFsc2V9"
            },
            "type": "public-key",
            "clientExtensionResults": {},
            "transports": [
                "nfc",
                "usb"
            ]
        }"""
        challenge = base64url_to_bytes(
            "8LBCiOY3q1cBZHFAWtS4AZZChzGphy67lK7I70zKi4yC7pgrQ2Pch7nAjLk1wq9greshIAsW2AjibhXjjI0TmQ"
        )
        rp_id = "localhost"
        expected_origin = "http://localhost:5000"

        verification = verify_registration_response(
            credential=credential,
            expected_challenge=challenge,
            expected_origin=expected_origin,
            expected_rp_id=rp_id,
        )

        assert verification.fmt == AttestationFormat.PACKED
        assert verification.credential_id == base64url_to_bytes(
            "syGQPDZRUYdb4m3rdWeyPaIMYlbmydGp1TP_33vE_lqJ3PHNyTd0iKsnKr5WjnCcBzcesZrDEfB_RBLFzU3k4w"
        )

    def test_verify_attestation_with_okp_public_key(self) -> None:
        credential = """{
            "id": "WlHiMqH6UhUs-d43z-aGlE3nsXuEOQpa9P9pwpqb4tmvtBMBfGvAV2wUrqBCDENjkkxd6kIRzZQKcluyOFlyW_vXVZSAEgod1xj-1QmFpuwyBVnlkQGefRbmUjbEt5iE4q3tdjy65EWIekO0SNjCQx3LxIJMzi25fgUkI9Y-gg0",
            "rawId": "WlHiMqH6UhUs-d43z-aGlE3nsXuEOQpa9P9pwpqb4tmvtBMBfGvAV2wUrqBCDENjkkxd6kIRzZQKcluyOFlyW_vXVZSAEgod1xj-1QmFpuwyBVnlkQGefRbmUjbEt5iE4q3tdjy65EWIekO0SNjCQx3LxIJMzi25fgUkI9Y-gg0",
            "response": {
                "attestationObject": "o2NmbXRmcGFja2VkZ2F0dFN0bXSjY2FsZyZjc2lnWEcwRQIgB3c0BvOsyJut14wj4XVWxXliicWZMZLDNF621Zz7h_8CIQCIOqRyWOazVhqlBrD-HL-83KWAvGZRgvW-4A9SE8BLFWN4NWOBWQLBMIICvTCCAaWgAwIBAgIEK_F8eDANBgkqhkiG9w0BAQsFADAuMSwwKgYDVQQDEyNZdWJpY28gVTJGIFJvb3QgQ0EgU2VyaWFsIDQ1NzIwMDYzMTAgFw0xNDA4MDEwMDAwMDBaGA8yMDUwMDkwNDAwMDAwMFowbjELMAkGA1UEBhMCU0UxEjAQBgNVBAoMCVl1YmljbyBBQjEiMCAGA1UECwwZQXV0aGVudGljYXRvciBBdHRlc3RhdGlvbjEnMCUGA1UEAwweWXViaWNvIFUyRiBFRSBTZXJpYWwgNzM3MjQ2MzI4MFkwEwYHKoZIzj0CAQYIKoZIzj0DAQcDQgAEdMLHhCPIcS6bSPJZWGb8cECuTN8H13fVha8Ek5nt-pI8vrSflxb59Vp4bDQlH8jzXj3oW1ZwUDjHC6EnGWB5i6NsMGowIgYJKwYBBAGCxAoCBBUxLjMuNi4xLjQuMS40MTQ4Mi4xLjcwEwYLKwYBBAGC5RwCAQEEBAMCAiQwIQYLKwYBBAGC5RwBAQQEEgQQxe9V_62aS5-1gK3rr-Am0DAMBgNVHRMBAf8EAjAAMA0GCSqGSIb3DQEBCwUAA4IBAQCLbpN2nXhNbunZANJxAn_Cd-S4JuZsObnUiLnLLS0FPWa01TY8F7oJ8bE-aFa4kTe6NQQfi8-yiZrQ8N-JL4f7gNdQPSrH-r3iFd4SvroDe1jaJO4J9LeiFjmRdcVa-5cqNF4G1fPCofvw9W4lKnObuPakr0x_icdVq1MXhYdUtQk6Zr5mBnc4FhN9qi7DXqLHD5G7ZFUmGwfIcD2-0m1f1mwQS8yRD5-_aDCf3vutwddoi3crtivzyromwbKklR4qHunJ75LGZLZA8pJ_mXnUQ6TTsgRqPvPXgQPbSyGMf2z_DIPbQqCD_Bmc4dj9o6LozheBdDtcZCAjSPTAd_uiaGF1dGhEYXRhWOFJlg3liA6MaHQ0Fw9kdmBbj-SuuaKGMseZXPO6gx2XY0EAAAACxe9V_62aS5-1gK3rr-Am0ACAWlHiMqH6UhUs-d43z-aGlE3nsXuEOQpa9P9pwpqb4tmvtBMBfGvAV2wUrqBCDENjkkxd6kIRzZQKcluyOFlyW_vXVZSAEgod1xj-1QmFpuwyBVnlkQGefRbmUjbEt5iE4q3tdjy65EWIekO0SNjCQx3LxIJMzi25fgUkI9Y-gg2kAQEDJyAGIVggnB_oUZDQU0esRlNPmjEO96aMDTgs34D8Dv31tAwhUZo",
                "clientDataJSON": "eyJ0eXBlIjoid2ViYXV0aG4uY3JlYXRlIiwiY2hhbGxlbmdlIjoiN0pVQmpXWkZkRm96dWx4YjcxRHZIa2gzUDZXS1VHNEVsVW83d0VrS2ljMkpFVEpNQUlLQ2Y3ckJFOVlrc0k1b05qRHpRNkhxaDZFNzNPeTZTUGNNbnciLCJvcmlnaW4iOiJodHRwOi8vbG9jYWxob3N0OjUwMDAiLCJjcm9zc09yaWdpbiI6ZmFsc2V9"
            },
            "type": "public-key",
            "clientExtensionResults": {},
            "transports": [
                "usb"
            ]
        }"""
        challenge = base64url_to_bytes(
            "7JUBjWZFdFozulxb71DvHkh3P6WKUG4ElUo7wEkKic2JETJMAIKCf7rBE9YksI5oNjDzQ6Hqh6E73Oy6SPcMnw"
        )
        rp_id = "localhost"
        expected_origin = "http://localhost:5000"

        verification = verify_registration_response(
            credential=credential,
            expected_challenge=challenge,
            expected_origin=expected_origin,
            expected_rp_id=rp_id,
        )

        assert verification.fmt == AttestationFormat.PACKED
        assert verification.credential_id == base64url_to_bytes(
            "WlHiMqH6UhUs-d43z-aGlE3nsXuEOQpa9P9pwpqb4tmvtBMBfGvAV2wUrqBCDENjkkxd6kIRzZQKcluyOFlyW_vXVZSAEgod1xj-1QmFpuwyBVnlkQGefRbmUjbEt5iE4q3tdjy65EWIekO0SNjCQx3LxIJMzi25fgUkI9Y-gg0"
        )
